/* 
 * WorldWide File System
 * Copyright (c) 1992,1993 Youki Kadobayashi
 * Copyright (c) 1992,1993 Osaka University
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that the following conditions
 * are met:
 * 1. Both the copyright notice and this permission notice appear in
 *    all copies of the software, derivative works or modified versions,
 *    and any portions thereof, and that both notices appear in
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Osaka University
 *      and its contributors.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE DEVELOPER ``AS IS'' AND OSAKA
 * UNIVERSITY DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
 * WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Osaka University requests users of this software to return to
 *
 *  Youki Kadobayashi
 *  Department of Information and Computer Sciences
 *  Osaka University, Toyonaka 560, Osaka, Japan
 *
 * any improvements or extensions that they make and grant Osaka
 * University the rights to redistribute these changes.
 */
/* CS filesystem */
static char *AtFSid = "$Header: cs.c[109.0] Wed Nov 24 03:47:04 1993 youki-k@is.aist-nara.ac.jp saved $";

#include <rpc/rpc.h>
#include "wfs.h"
#include "global.h"
#include "util.h"

/* translation of WFS semantics to CS semantics */
static cs_stat cserr[] = {
	CS_OK,		/* WF_REP_OK: no error */
	CSERR_OTHER,	/* WF_ERR_PERM: Not owner *//* ROFS */
	CSERR_NOENT,	/* WF_ERR_NOENT: No such entity */
	CSERR_OTHER,	/* WF_ERR_IO: I/O error *//* ROFS */
	CSERR_OTHER,	/* WF_ERR_NXIO: No such device or address *//* NODEV */
	CSERR_ACCES,	/* WF_ERR_ACCES: Permission denied */
	CSERR_OTHER,	/* WF_ERR_EXIST: Given resource exists *//* ROFS */
	CSERR_OTHER,	/* WF_ERR_NODEV: No such device *//* NODEV */
	CSERR_NOTDIR,	/* WF_ERR_NOTDIR: Not a directory*/
	CSERR_ISDIR,	/* WF_ERR_ISDIR: Is a directory */
	CSERR_FBIG,	/* WF_ERR_FBIG: Resource too large */
	CSERR_NOSPC,	/* WF_ERR_NOSPC: No space left on device */
	CSERR_ROFS,	/* WF_ERR_ROFS: Read-only file system */
	CSERR_OTHER,	/* WF_ERR_NAMETOOLONG: Resource name too long *//*XXX*/
	CSERR_OTHER,	/* WF_ERR_NOTEMPTY: Directory not empty *//* ROFS */
	CSERR_OTHER,	/* WF_ERR_DQUOT: Disc quota exceeded *//* ROFS */
	CSERR_STALE,	/* WF_ERR_STALE: Stale file handle */
	CSERR_OTHER,	/* WF_ERR_WFLUSH: Write cache flushed *//* ROFS */
	CS_WAIT,	/* WF_REP_WAIT: Operation would block */
	CSERR_ALREADY,	/* WF_ERR_ALREADY: Operation already in progress */
	CSERR_PROTO,	/* WF_ERR_PROTO: Protocol not available */
	CSERR_NDOWN,	/* WF_ERR_NDOWN: Network is down */
	CSERR_NUNREACH,	/* WF_ERR_NUNREACH: Network is unreachable */
	CSERR_XABORT,	/* WF_ERR_XABORT: Transfer aborted */
	CSERR_TIMEOUT,	/* WF_ERR_TIMEOUT: Connection timed out */
	CSERR_REFUSED,	/* WF_ERR_REFUSED: Connection refused */
	CSERR_HDOWN,	/* WF_ERR_HDOWN: Host is down */
	CSERR_HUNREACH,	/* WF_ERR_HUNREACH: No route to host */
	CSERR_NQUOT,	/* WF_ERR_NQUOT: Network quota exceeded */
};

int
cs_namei(pname, fhp)
char *pname;
wf_fh *fhp;	/* result */
{
	int ret;
	ret = root_name2fh(pname, fhp);
	return cserr[ret];
}

long
cs_error(msg)
char *msg;
{
	long err_context;

	return err_context;
}

extern wf_vol *root_vol;
extern wf_dir *root_dir;

void
cs_mount(c)
wf_thrd *c;
{
	wf_file *filep;
	filep = file_findname(root_dir, c->fname);
	if (filep) {
		/* already mounted */
		(*c->reply)(c, WF_REP_OK);
		return;
	}

	/* mount it */
	c->parent_volp = c->child_volp = root_vol;
	c->fhp = ALLOC(wf_fh);
	c->fhp->parent_vol = c->fhp->child_vol = WF_ROOT_VOL_ID;
	c->fhp->dir_id = c->fhp->file_id = WF_ROOT_DIR_ID;
	c->dotdotp = root_dir;
	root_lookup_miss(c);
}

void
cs_mount_done(c, status)
wf_thrd *c;
int status;
{
	static cs_res res;

	if (status == WF_REP_WAIT) {
		return;
	}
	res.status = cserr[status];
	svc_sendreply(c->rqstp->rq_xprt, xdr_cs_res, &res);
	if (c->fhp) FREE(c->fhp);
	if (! (c->flag & WF_THREAD_BUSY)) thrd_free(c);
}

int
cs_umount(vol)
char *vol;
{
	wf_file *filep;
	wf_vol *volp;

	filep = file_findname(root_dir, vol);
	if (! filep) {
		return CSERR_NOENT;
	}
	q_remove((qelem *) filep);
	volp = vol_findname(".", vol);		/* XXX must support subdirs */
	if (! volp) {
		return CSERR_NOENT;
	}
	vol_free(volp);
	return CS_OK;
}

void
cs_lookup(rqstp, volp, fhp, fname)
svc_req *rqstp;
wf_vol *volp;
wf_fh *fhp;
char *fname;
{
}

void
cs_readdir(rqstp, volp, fhp, cookie, count)
svc_req *rqstp;
wf_vol *volp;
wf_fh *fhp;
long cookie;
u_int count;
{
}

void
cs_read(rqstp, volp, fhp, offset, count, totalcount)
svc_req *rqstp;
wf_vol *volp;
wf_fh *fhp;
u_int offset, count, totalcount;
{
}

void
cs_getuda(rqstp, fhp, attrname)
svc_req *rqstp;
wf_fh *fhp;
char *attrname;
{
}

void
cs_geterr(rqstp, err_context)
svc_req *rqstp;
long err_context;
{
}
