/* 
 * WorldWide File System
 * Copyright (c) 1992,1993 Youki Kadobayashi
 * Copyright (c) 1992,1993 Osaka University
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that the following conditions
 * are met:
 * 1. Both the copyright notice and this permission notice appear in
 *    all copies of the software, derivative works or modified versions,
 *    and any portions thereof, and that both notices appear in
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Osaka University
 *      and its contributors.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE DEVELOPER ``AS IS'' AND OSAKA
 * UNIVERSITY DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
 * WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Osaka University requests users of this software to return to
 *
 *  Youki Kadobayashi
 *  Department of Information and Computer Sciences
 *  Osaka University, Toyonaka 560, Osaka, Japan
 *
 * any improvements or extensions that they make and grant Osaka
 * University the rights to redistribute these changes.
 */
/*
 * RPC stubs for CS (WWFS Cluster Server) protocol.
 *
 * Darn. it would be far easier if I could write filesystems in Haskell.
 */
static char *AtFSid = "$Header: cs_subr.c[109.0] Wed Nov 24 03:47:06 1993 youki-k@is.aist-nara.ac.jp saved $";

#include "wfs.h"
#include "global.h"
#include "util.h"

void *
cs_proc_null_1(argp, rqstp)	/* CS RPC keepalive */
void *argp;
svc_req *rqstp;
{
	static char res;
	return (void *) &res;
}

int *
cs_proc_getport_1(argp, rqstp)	/* XXX get NFS port - obsolete */
void *argp;
svc_req *rqstp;
{
	static int res;

#ifdef DEBUG_CSPROC
	dlog("cs_proc_getport_1");
#endif
	bzero((char *)&res, sizeof(res));
	res = cs_nfsport;
	return &res;
}

cs_fhres *
cs_proc_register_1(argp, rqstp)
void *argp;
svc_req *rqstp;
{
	static cs_fhres res;
	static wf_fh *fhp = (wf_fh *) &res.cs_fhres_u.file;

	/* Establish relationship with csd by initial filehandle. */
	bzero((char *)&res, sizeof(res));
	res.status = CS_OK;
	fhp->world_id = cs_world;
	fhp->parent_vol = fhp->child_vol = WF_ROOT_VOL_ID;
	fhp->dir_id = fhp->file_id = WF_ROOT_DIR_ID;
	return &res;
}

cs_res *
cs_proc_unregister_1(argp, rqstp)
void *argp;
svc_req *rqstp;
{
	static cs_res res;

	/* Purge relationship with csd. Just for "showmount" function. */
	res.status = CS_OK;
	bzero((char *)&res, sizeof(res));
	return &res;
}

static int
is_cluster_root(rqstp)
svc_req *rqstp;
{
	sockaddr_in *sin;
	struct authunix_parms *unix_cred;

	sin = svc_getcaller(rqstp->rq_xprt);
	if (rqstp->rq_cred.oa_flavor == AUTH_UNIX) {
		unix_cred = (struct authunix_parms *) rqstp->rq_clntcred;
		if (unix_cred->aup_uid == 0
		    && bcmp(&sin->sin_addr, &cs_ipaddr,
			    sizeof(struct in_addr)) == 0) {
			return 1;
		}
	}
	return 0;
}

/* volname -> Bool */
cs_res *
cs_proc_mount_1(argp, rqstp)
cs_volargs *argp;
svc_req *rqstp;
{
	static cs_res res;
	wf_thrd *c;

#ifdef DEBUG_CSPROC
	dlog("cs_proc_mount_1");
#endif
	if (!is_cluster_root(rqstp)) {
		bzero((char *)&res, sizeof(res));
		res.status = CSERR_ACCES;
		return &res;
	}
	c = thrd_alloc();
	c->rqstp = CLONE(rqstp);
	c->fname = argp->vol;
	c->reply = cs_mount_done;
	cs_mount(c);
	return (cs_res *)0;
}

/* volname -> Bool */
cs_res *
cs_proc_umount_1(argp, rqstp)
cs_volargs *argp;
svc_req *rqstp;
{
	static cs_res res;

#ifdef DEBUG_CSPROC
	dlog("cs_proc_umount_1");
#endif
	bzero((char *)&res, sizeof(res));
	if (!is_cluster_root(rqstp)) {
		res.status = CSERR_ACCES;
		return &res;
	}
	res.status = cs_umount(argp->vol);
	return &res;
}

/* pathname -> filename -> [nfs_fh, fattr] */
cs_diropres *
cs_proc_lookup_1(argp, rqstp)
cs_diropargs *argp;
svc_req *rqstp;
{
	static cs_diropres res;
	wf_fh fh;
	wf_vol *volp;
	int ret;

#ifdef DEBUG_CSPROC
	dlog("cs_proc_lookup_1");
#endif

	ret = cs_namei(argp->pname, &fh);
	if (ret != CS_OK) {
		bzero(&res, sizeof(res));
		res.status = ret;
		return &res;
	}
	volp = vol_findid(fh.child_vol);
	if (volp == NULL) {
		bzero(&res, sizeof(res));
		res.status = CSERR_STALE;
		return &res;
	}
	cs_lookup(rqstp, volp, &fh, argp->fname);
	return (cs_diropres *)0;
}

/* nfs_fh -> cookie -> count -> [[file id, filename, fattr]] */
cs_readdirres *
cs_proc_readdir_1(argp, rqstp)
cs_readdirargs *argp;
svc_req *rqstp;
{
	static cs_readdirres res;
	wf_fh *fhp = (wf_fh *) &argp->dir;
	wf_vol *volp;

#ifdef DEBUG_CSPROC
	dlog("cs_proc_readdir_1");
#endif
	bzero(&res, sizeof(res));
	/* auth */
	if (fhp->world_id != cs_world) {
		res.status = CSERR_ACCES;
		return &res;
	}

	volp = vol_findid(fhp->child_vol);
	if (volp == NULL) {
		res.status = CSERR_STALE;
		return &res;
	}
	cs_readdir(rqstp, volp, fhp, argp->cookie, argp->count);
	return (cs_readdirres *)0;
}

cs_readres *
cs_proc_read_1(argp, rqstp)
cs_readargs *argp;
svc_req *rqstp;
{
	static cs_readres res;
	wf_fh *fhp = (wf_fh *) &argp->file;
	wf_vol *volp;

#ifdef DEBUG_CSPROC
	dlog("cs_proc_read_1");
#endif
	bzero(&res, sizeof(res));
	/* auth */
	if (fhp->world_id != cs_world) {
		res.status = CSERR_ACCES;
		return &res;
	}

	volp = vol_findid(fhp->child_vol);
	if (volp == NULL) {
		res.status = CSERR_STALE;
		return &res;
	}
	cs_read(rqstp, volp, fhp, argp->offset, argp->count, argp->totalcount);
	return (cs_readres *)0;
}

cs_volres *
cs_proc_getvol_1(argp, rqstp)
cs_volargs *argp;
svc_req *rqstp;
{
	static cs_volres res;
	cs_volokres *p;
	wf_vol *volp;

#ifdef DEBUG_CSPROC
	dlog("cs_proc_getvol_1");
#endif
	bzero((char *)&res, sizeof(res));
	volp = vol_findname(".", argp->vol);	/* XXX must support subdirs */
	if (volp == NULL) {
		res.status = CSERR_NOENT;
		return &res;
	}
	p = &res.cs_volres_u.volok;
	p->volume_id = volp->id;
	p->bytes_to_server = volp->stats.bytes_to_server;
	p->bytes_from_server = volp->stats.bytes_from_server;
	p->bytes_to_client = volp->stats.bytes_from_server;
	p->bytes_from_client = volp->stats.bytes_from_client;
	p->n_request_readdir = volp->stats.n_request_readdir;
	p->n_request_lookup = volp->stats.n_request_lookup;
	p->n_request_read = volp->stats.n_request_read;
	p->n_request_readdir_miss = volp->stats.n_request_readdir_miss;
	p->n_request_lookup_miss = volp->stats.n_request_lookup_miss;
	p->n_request_read_miss = volp->stats.n_request_read_miss;
	return &res;
}

cs_xferres *
cs_proc_getxfer_1(fhp, rqstp)
wf_fh *fhp;
svc_req *rqstp;
{
	static cs_xferres res;
	wf_vol *volp;
	wf_dir *dirp;
	wf_file *filep;
	wf_thrd *thrdp;
	sockaddr_in *sin;

#ifdef DEBUG_CSPROC
	dlog("cs_proc_getxfer_1");
#endif
	bzero(&res, sizeof(res));
	volp = vol_findid(fhp->child_vol);
	if (!volp) {
		res.status = CSERR_STALE;
		return &res;
	}
	dirp = dir_findcache(volp, fhp->dir_id);
	if (!dirp) {
		res.status = CSERR_STALE;
		return &res;
	}
	filep = file_findid(dirp, fhp->child_vol, fhp->file_id);
	if (!filep) {
		res.status = CSERR_STALE;
		return &res;
	}
	if (!filep->thrdp) {
		/* not transferring this file */
		res.status = CSERR_NOENT;
		return &res;
	}
	thrdp = filep->thrdp;
	res.status = CS_OK;
	if (thrdp->rqstp->rq_cred.oa_flavor == AUTH_UNIX) {
                struct authunix_parms *unix_cred;
		unix_cred = (struct authunix_parms *) thrdp->rqstp->rq_clntcred;
		res.cs_xferres_u.xfer.uid = unix_cred->aup_uid;
		res.cs_xferres_u.xfer.gid = unix_cred->aup_gid;
	} else {
		res.cs_xferres_u.xfer.uid = 54321;	/* TNFS? yittch. */
		res.cs_xferres_u.xfer.gid = 54321;
	}
	sin = svc_getcaller(thrdp->rqstp->rq_xprt);
	res.cs_xferres_u.xfer.ip_address = sin->sin_addr.s_addr;
	res.cs_xferres_u.xfer.bytes = thrdp->bytes;
	res.cs_xferres_u.xfer.idle = thrdp->idle;
	res.cs_xferres_u.xfer.server.server_len = strlen(thrdp->cp->srv->name);
	res.cs_xferres_u.xfer.server.server_val = thrdp->cp->srv->name;
	return &res;
}

cs_udares *
cs_proc_getuda_1(argp, rqstp)
cs_udaargs *argp;
svc_req *rqstp;
{
	static cs_udares res;

#ifdef DEBUG_CSPROC
	dlog("cs_proc_getuda_1");
#endif
	bzero((char *)&res, sizeof(res));
	cs_getuda(rqstp, (wf_fh *) &argp->file, argp->attrname);
	return &res;
}

cs_geterrres *
cs_proc_geterr_1(argp, rqstp)
cs_geterrargs *argp;
svc_req *rqstp;
{
	static cs_geterrres res;

#ifdef DEBUG_CSPROC
	dlog("cs_proc_geterr_1");
#endif
	bzero((char *)&res, sizeof(res));
	cs_geterr(rqstp, argp->err_context);
	return &res;
}

