/* 
 * WorldWide File System
 * Copyright (c) 1992,1993 Youki Kadobayashi
 * Copyright (c) 1992,1993 Osaka University
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that the following conditions
 * are met:
 * 1. Both the copyright notice and this permission notice appear in
 *    all copies of the software, derivative works or modified versions,
 *    and any portions thereof, and that both notices appear in
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Osaka University
 *      and its contributors.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE DEVELOPER ``AS IS'' AND OSAKA
 * UNIVERSITY DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
 * WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Osaka University requests users of this software to return to
 *
 *  Youki Kadobayashi
 *  Department of Information and Computer Sciences
 *  Osaka University, Toyonaka 560, Osaka, Japan
 *
 * any improvements or extensions that they make and grant Osaka
 * University the rights to redistribute these changes.
 */
/* NFS over FTP */
static char *AtFSid = "$Header: ftp.c[109.0] Wed Nov 24 03:47:09 1993 youki-k@is.aist-nara.ac.jp saved $";

#include <sys/ioctl.h>		/* for FIONBIO */
#include <ctype.h>		/* for isdigit */
#include <string.h>		/* for strchr */
#include "wfs.h"
#include "util.h"
#include "global.h"

/* public functions  */
void ftp_getuda(), ftp_keepalive(), ftp_close(), ftp_shutdown();

wf_proto proto_ftp = {
	"FTP", cmfs_getattr, err_getattr_miss,
	cmfs_lookup, err_lookup_miss, cmfs_readlink,
	cmfs_read, ftp_read_miss, cmfs_readdir, ftp_readdir_miss,
	ftp_getuda, ftp_keepalive, ftp_close, ftp_shutdown, 0
};

/* debug */
wf_req *cs_ftp_debug;

void
ftp_getuda()
{
	/* XXX: not used? */
}

void
ftp_keepalive()
{
	/* XXX: TCP keepalive could be used */
}

void
ftp_close(cp)
wf_conn *cp;
{
#ifdef VERBOSE
	dlog("closing FTP connection to %s", cp->srv->name);
#endif
	if (cp->thrdp && cs_state < Finishing) {
		dlog("transfer aborted");
	}
	send(cp->id, "QUIT\r\n", 6, 0);
	so_clear(cp->id);
	so_unregister(cp->id, WF_SO_ALL);
	close(cp->id);
	cp->thrdp = (wf_thrd *)0;
}

void
ftp_shutdown(cp, howto)
wf_conn *cp;
int howto;
{
	so_clear(cp->id);
	so_unregister(cp->id, WF_SO_ALL);
	shutdown(cp->id, howto);
}

/* connection */
int
ftp_conn_open(c)
wf_thrd *c;	/* cp */
{
	int so, opt;
	static void ftp_conn_nbio();

	so = c->cp->id = socket(AF_INET, SOCK_STREAM, 0);
	if (so < 0) {
		errno_diag();
		return -1;
	}
#ifdef VERBOSE
	talk(c->client, "opening FTP connection to %s", c->cp->srv->name);
	dlog("opening FTP connection to %s", c->cp->srv->name);
#endif

	opt = 1;
	ioctl(so, FIONBIO, &opt);
#ifdef HZ
	opt = 3 * HZ;				/* 3sec */
	setsockopt(so, SOL_SOCKET, SO_SNDTIMEO, (char *)&opt, sizeof(int));
	setsockopt(so, SOL_SOCKET, SO_RCVTIMEO, (char *)&opt, sizeof(int));
#endif
	c->step = FTPS_CONN_INIT;
	ftp_conn_nbio(c);
	return so;
}

static void
ftp_conn_fail(c)
wf_thrd *c;
{
	c->flag |= WF_THREAD_ERROR;
	c->step = FTPS_CONN_FAIL;
	wakeup(&c->cp);
}

static void
ftp_conn_nbio(c)
wf_thrd *c;
{
	int so = c->cp->id;
	int ret;
	static wf_thrd x;

	switch (c->step) {
	case FTPS_CONN_INIT:
		c->give_up = ftp_conn_fail;
		c->closure = (void *)c;
		c->cp->thrdp = &x;	/* fake */
		connect(so, (struct sockaddr *)&c->cp->srv->sin,
			sizeof (sockaddr_in));
		so_register(so, WF_SO_READ|WF_SO_WRITE);
		so_callback(so, ftp_conn_nbio, c);
		c->step = FTPS_CONN_USER;
		return;
	case FTPS_CONN_USER:
		so_unregister(so, WF_SO_WRITE);
		ret = tcp_send(so, "USER anonymous\r\n");
		if (cs_ftp_debug) {
			req_send(cs_ftp_debug, "USER anonymous\r\n");
		}
		if (ret < 0) {
			c->flag |= WF_THREAD_ERROR;
			srv_flag(c->cp->srv, WF_SRV_REFUSED);
			break;
		}
		c->step = FTPS_CONN_PASS;
		return;
	case FTPS_CONN_PASS:
		ret = ftp_recv_status(c->cp);
		if (ret == 331) {
			tcp_send(so, "PASS wwfs@%s.%s\r\n",
				 cs_hostname, cs_domain);
			if (cs_ftp_debug) {
				req_send(cs_ftp_debug, "PASS wwfs@%s.%s\r\n",
					 cs_hostname, cs_domain);
			}
			c->step = FTPS_CONN_TYPE;
		} else if (ret == 230) {
			/* "230 User logged in, proceed." */
			goto bypass_pass;
		} else if (ret == 0 || ret < 400) {
			/* expect more replies */
			c->step = FTPS_CONN_PASS;
		} else {
			/* "530 User anonymous unknown" or somesuch */
			ftp_reply_diag(ret);
			c->flag |= WF_THREAD_ERROR;
			srv_flag(c->cp->srv, WF_SRV_REFUSED);
			break;
		}
		return;
	case FTPS_CONN_TYPE:
		ret = ftp_recv_status(c->cp);
		if (ret == 230) {
		bypass_pass:
			tcp_send(so, "TYPE I\r\n");
			if (cs_ftp_debug) {
				req_send(cs_ftp_debug, "TYPE I\r\n");
			}
			c->step = FTPS_CONN_FIN;
		} else if (ret == 0 || ret < 400) {
			/* expect more replies */
			c->step = FTPS_CONN_TYPE;
		} else {
			/* e.g., "530 Too many users connected currently" */
			ftp_reply_diag(ret);
			c->flag |= WF_THREAD_ERROR;
			srv_flag(c->cp->srv, WF_SRV_REFUSED);
			break;
		}
		return;
	case FTPS_CONN_FIN:
		ret = ftp_recv_status(c->cp);
		if (ret == 200) {
			so_clear(so);
			c->cp->thrdp = (wf_thrd *)0;
			so_callback(so, ftp_recv_junk, (void *)so);
			break;
		} else if (ret == 0 || ret < 400) {
			/* expect more replies */
			c->step = FTPS_CONN_FIN;
		} else {
			ftp_reply_diag(ret);
		}
		return;
	}
	c->step = FTPS_PROMPT;
	wakeup(&c->cp);
}

int
ftp_recv_status(cp)
wf_conn *cp;
{
	long navail;
	int nread;
	int ret, so, status = 0;
	char *bufp, *p, *q;

	so = cp->id;
	bufp = so_getbuf(so);
	ioctl(so, FIONREAD, &navail);
	nread = MIN(navail, NFS_MAXDATA);
	if (nread == 0) {
		if (cp->thrdp) cp->thrdp->flag |= WF_THREAD_ERROR;
		srv_flag(cp->srv, WF_SRV_XABORT);
		return 426;	/* connection closed */
	}
	ret = recv(so, bufp, nread, MSG_PEEK);
	if (ret < 0) {
		errno_diag();
		if (cp->thrdp) cp->thrdp->flag |= WF_THREAD_ERROR;
		return 451;	/* local error */
	}
	bufp[nread] = '\0';
	if ((q = strrchr(bufp, '\n')) == NULL)
		return 0;	/* too fast? */
	recv(so, bufp, q-bufp+1, 0);
	q[1] = '\0';
	p = bufp;
	while (*p) {
		q = strchr(p, '\n');
		if (isdigit(*p) && isspace(p[3])) {
			status = atoi(p);
			if (status > 400 && cp->thrdp)
				cp->thrdp->flag |= WF_THREAD_ERROR;
		}
		p = q+1;
	}
	return status;
}

void
ftp_recv_junk(so)	/* firewall */
int so;
{
	long navail;
	int nread, ret, optlen;
	char *bufp;

	bufp = so_getbuf(so);
	ioctl(so, FIONREAD, &navail);
	nread = MIN(navail, NFS_MAXDATA-64);
	if (nread == 0) {
		optlen = sizeof(int);
		getsockopt(so, SOL_SOCKET, SO_ERROR, &ret, &optlen);
		if (ret && ret < sys_nerr)
			dlog("ftp_recv_junk: %s", sys_errlist[ret]);
		conn_dropped(so);
		return;
	}
	ret = recv(so, bufp, nread, 0);
	if (ret < 0) {
		errno_diag();
		return;
	}
	bufp[nread] = '\0';
	dlog("ftp_recv_junk(%d): %s", so, bufp);
}

void
ftp_reply_diag(ret)
int ret;
{
	/* See RFC959 for details */
	char *p;

	switch (ret) {
	case 110:
		p = "110 Restart marker reply.";
		break;
	case 120:
		p = "120 Service ready in nnn minutes.";
		break;
	case 125:
		p = "125 Data connection already open; transfer starting.";
		break;
	case 150:
		p = "150 File status okay; about to open data connection.";
		break;
	case 200:
		p = "200 Command okay.";
		break;
	case 202:
		p = "202 Command not implemented, superflous at this state.";
		break;
	case 211:
		p = "211 System status, or system help reply.";
		break;
	case 212:
		p = "212 Directory status.";
		break;
	case 213:
		p = "213 File status.";
		break;
	case 214:
		p = "214 Help message.";
		break;
	case 215:
		p = "215 NAME system type.";
		break;
	case 220:
		p = "220 Service ready for new user.";
		break;
	case 221:
		p = "221 Service closing control connection.";
		break;
	case 225:
		p = "225 Data connection open; no transfer in progress.";
		break;
	case 226:
		p = "226 Closing data connection.";
		break;
	case 227:
		p = "227 Entering passive mode.";
		break;
	case 230:
		p = "230 User logged in, proceed.";
		break;
	case 250:
		p = "250 Requested file action okay, completed.";
		break;
	case 257:
		p = "257 PATHNAME created.";
		break;
	case 331:
		p = "331 User name okay, need password.";
		break;
	case 332:
		p = "332 Need account for login.";
		break;
	case 350:
		p = "350 Requested file action pending further information.";
		break;
	case 421:
		p = "421 Service not available, closing control connection.";
		break;
	case 425:
		p = "425 Can't open data connection.";
		break;
	case 426:
		p = "426 Connection closed; transfer aborted.";
		break;
	case 450:
		p = "450 Requested file action not taken.";
		break;
	case 451:
		p = "451 Requested action aborted: local error in processing.";
		break;
	case 452:
		p = "452 Requested action not taken. Insufficient storage space in system.";
		break;
	case 500:
		p = "500 Syntax error, command unrecognized.";
		break;
	case 501:
		p = "501 Syntax error in parameters or arguments.";
		break;
	case 502:
		p = "502 Command not implemented.";
		break;
	case 503:
		p = "503 Bad sequence of commands.";
		break;
	case 504:
		p = "504 Command not implemented for that parameter.";
		break;
	case 530:
		p = "530 Not logged in.";
		break;
	case 532:
		p = "532 Need account for storing files.";
		break;
	case 550:
		p = "550 Requested action not taken. File unavailable.";
		break;
	case 551:
		p = "551 Requested action aborted: page type unknown.";
		break;
	case 552:
		p = "552 Requested file action aborted. Exceeded storage allocation.";
		break;
	case 553:
		p = "553 Requested action not taken. File name not allowed.";
		break;
	default:
		dlog("ftp_reply_diag: unknown reply code %d", ret);
		return;
	}
	dlog(p);
}

