/* 
 * WorldWide File System
 * Copyright (c) 1992,1993 Youki Kadobayashi
 * Copyright (c) 1992,1993 Osaka University
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that the following conditions
 * are met:
 * 1. Both the copyright notice and this permission notice appear in
 *    all copies of the software, derivative works or modified versions,
 *    and any portions thereof, and that both notices appear in
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Osaka University
 *      and its contributors.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE DEVELOPER ``AS IS'' AND OSAKA
 * UNIVERSITY DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
 * WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Osaka University requests users of this software to return to
 *
 *  Youki Kadobayashi
 *  Department of Information and Computer Sciences
 *  Osaka University, Toyonaka 560, Osaka, Japan
 *
 * any improvements or extensions that they make and grant Osaka
 * University the rights to redistribute these changes.
 */
/* NFS over FTP */
static char *AtFSid = "$Header: ftp_list.c[109.0] Wed Nov 24 03:47:09 1993 youki-k@is.aist-nara.ac.jp saved $";

#include <sys/types.h>
#include <sys/ioctl.h>		/* for FIONREAD */
#include <ctype.h>
#include <string.h>		/* for strchr */
#include "wfs.h"
#include "util.h"
#include "global.h"

static void
ftp_readdir_lock(c)
wf_thrd *c;
{
	c->filep->thrdp = c;
	++c->dotdotp->n_xfer;
}

static void
ftp_readdir_unlock(c)
wf_thrd *c;
{
	c->filep->thrdp = (wf_thrd *)0;
	warn_negative(--c->dotdotp->n_xfer);
	if (c->cp)
		c->cp->thrdp = (wf_thrd *)0;
}

static void ftp_readdir_2(), ftp_readdir_3(), ftp_readdir_fail();
void
ftp_readdir_miss(c)
wf_thrd *c;
/* in: rqstp, volp, filep, fhp, cookie, count, dotdotp, olddirp */
/* out: cp, dirp, flag */
{
	wf_file *p;

	ftp_readdir_lock(c);
	if (c->fhp->child_dir > WF_VOL_BASE) {
		if (c->olddirp) {
			c->dirp = dir_new(c->child_volp, c->fhp->child_dir,
					  c->dotdotp->name, "");
		} else {
			c->dirp = vol_newroot(c->child_volp);
		}
	} else {
		c->dirp = dir_new(c->child_volp, c->fhp->child_dir,
				  c->dotdotp->name, c->filep->name);
	}
	p = dir_adddir(c->dirp, ".", c->child_volp->id, c->dirp->id);
	attr_touch(&p->fattr);
	p = dir_adddir(c->dirp, "..", c->parent_volp->id, c->dotdotp->id);
	attr_touch(&p->fattr);

	c->cp = vol_getconn(c->child_volp);
	if (c->cp == (wf_conn *)0) {
		if ((c->child_volp->flag & WF_VOL_XCONN)
		    || (c->cp = vol_newconn(c->child_volp)) == (wf_conn *)0) {
			/* exceeded resource limit */
			ftp_readdir_unlock(c);
			(*c->reply)(c, WF_ERR_NQUOT);
			return;
		}
		assert(c->cp != (wf_conn *)0);
		/* new connection needed */
		sched_task(ftp_readdir_2, c, &c->cp);
		thrd_watch(c);
		c->flag |= WF_THREAD_BUSY;
		ftp_conn_open(c);
		return;
	}
	c->flag |= WF_THREAD_BUSY;
	ftp_readdir_2(c);
}

static void
ftp_readdir_2(c)
wf_thrd *c;
{
	int ret;

	c->idle = 0;
	switch (c->step) {
	case FTPS_CONN_FAIL:
		ftp_readdir_fail(c);
		return;
	case FTPS_PROMPT:
#ifdef VERBOSE
		talk(c->client, "listing directory %s", c->dirp->name);
		dlog("listing directory %s", c->dirp->name);
#endif
		c->cp->thrdp = c;
		c->flag |= WF_THREAD_RISKY;
		c->give_up = ftp_readdir_fail;
		c->closure = (void *)c;
		ret = tcp_send(c->cp->id, "STAT %s/%s%s/\r\n",
			       c->cp->srv->variant & WF_FTP_FOLLOW ? "-L ": "",
			       c->cp->info, c->dirp->name);
		if (cs_ftp_debug) {
			req_send(cs_ftp_debug, "STAT /%s%s/\r\n",
				 c->cp->info, c->dirp->name);
		}
		if (ret < 0) {
			ftp_readdir_fail(c);
			return;
		}
		c->step = FTPS_STAT_RECV;
		so_clear(c->cp->id);
		so_callback(c->cp->id, ftp_readdir_2, c);
		return;
	case FTPS_STAT_RECV:
		so_clear(c->cp->id);
		/* now receive listings. */
		c->step = FTPS_STAT_RECVINIT;
		sched_task(ftp_readdir_3, c, &c->cp->id);
		ftp_recv_list(c);
		return;
	}
}

#ifdef TRACE
static void
ftp_readdir_trace(c)
wf_thrd *c;
{
	wf_trace *tracep;
	tracep = ALLOC(wf_trace);
	tracep->time = gettime();
	tracep->volume = c->child_volp->id;
	tracep->dir = c->dirp->id;
	tracep->op = WF_OP_READDIR;
	tracep->hit = WF_CACHE_READY;
	trace_event(c, tracep);
}
#endif

static void
ftp_readdir_3(c)
wf_thrd *c;
{
	ftp_readdir_unlock(c);
	vol_releaseconn(c->child_volp);
	c->flag &= ~WF_THREAD_BUSY;

#ifdef TRACE
	ftp_readdir_trace(c);
#endif

	if (c->flag & WF_THREAD_ERROR) {
#ifdef VERBOSE
		talk(c->client, "FTP readdir failed for volume \"%s\" directory \"%s\"",
		     c->child_volp->name, c->dirp->name);
		dlog("FTP readdir failed for volume \"%s\" directory \"%s\"",
		     c->child_volp->name, c->dirp->name);
#endif
		(*c->reply)(c, WF_ERR_IO);
	} else {
		wf_file *f;

		f = file_findname(c->dirp, ".");
		c->filep->fattr.nlink = f->fattr.nlink;
		dir_save(c->child_volp, c->dirp, c->olddirp);
		(*c->reply)(c, WF_REP_OK);
	}
}

static void
ftp_readdir_fail(c)
wf_thrd *c;
{
	wf_conn *cp = c->cp;
	c->flag |= WF_THREAD_ERROR;
	ftp_readdir_3(c);
	conn_shutdown(cp, 2);
}

void
ftp_recv_list(c)
wf_thrd *c;
{
	long navail;
	int nread;
	char *q;
	int ret, optlen;

	c->idle = 0;
	switch (c->step) {
	case FTPS_STAT_RECVINIT:
		/* initialize */
		so_callback(c->cp->id, ftp_recv_list, c);
		c->buf = so_getbuf(c->cp->id);
	case FTPS_STAT_RECVLOOP:
		c->step = FTPS_STAT_RECVLOOP;
		ioctl(c->cp->id, FIONREAD, &navail);
		nread = MIN(navail, NFS_MAXDATA);
		if (nread == 0) {
			optlen = sizeof(int);
			getsockopt(c->cp->id, SOL_SOCKET, SO_ERROR,
				   &ret, &optlen);
			if (ret && ret < sys_nerr)
				dlog("ftp_recv_list: %s", sys_errlist[ret]);
			c->flag |= WF_THREAD_ERROR;
			break;
		}
		ret = recv(c->cp->id, c->buf, nread, MSG_PEEK);
		if (ret < 0) {
			errno_diag();
			c->flag |= WF_THREAD_ERROR;
			break;
		}
		c->buf[nread] = '\0';
		if ((q = strrchr(c->buf, '\n')) == NULL) return;
		nread = q - c->buf + 1;
		recv(c->cp->id, c->buf, nread, 0);
		c->cp->stats.bytes_from_server += nread;
		c->child_volp->stats.bytes_from_server += nread;
		q[1] = '\0';
		ret = ftp_parse_list(c->buf, c->child_volp,
					c->dirp, c->olddirp);
		if (ret == 211)
			break;
		if (ret > 400) {
			c->flag |= WF_THREAD_ERROR;
			break;
		}
		return;
	}
	so_clear(c->cp->id);
	so_callback(c->cp->id, ftp_recv_junk, (void *)c->cp->id);
	wakeup(&c->cp->id);
}

/* parse "LIST" output */
int
ftp_parse_list(buftop, volp, dirp, olddirp)
char *buftop;
wf_vol *volp;
wf_dir *dirp, *olddirp;
{
	char *p, *bufp = buftop;
	struct tm tm;
	int i, status = 0;

	wf_file *f;
	char *f_name, *f_link;
	ftype f_type;
	u_int f_nlink, f_mode, f_uid, f_gid, f_id, f_fsid;
	long f_size, f_time;

	/* ^[\-ld][\-rwxs]{9} \d+ [^ ]+ [^ ]+ \d+ [a-zA-Z]{3} \d+ (\d\d:\d\d|\d\d\d\d) [^ ]+ */
	while (1) {
		if (*bufp == '\0') {
			return status;
		}
		if (strchr("-bcdl", *bufp)) {
			/* /^[\-bcdl]/ -- file type */
			f_link = 0, f_mode = 0, f_id = 0;
			switch (*bufp) {
			case 'b':
				f_type = NFBLK;
				break;
			case 'c':
				f_type = NFCHR;
				break;
			case 'd':
				f_type = NFDIR;
				break;
			case 'l':
				f_type = NFLNK;
				break;
			case '-':
				f_type = NFREG;
				break;
			default:
				dlog("ftp_parse_list: unknown file type %c", *bufp);
				return 550;
			}
			++bufp;
		} else {
			if (isdigit(*bufp)) {
				/* /^\d{3}\s/ -- status */
				/* ignore /^\d{3}\-/ */
				if (isspace(bufp[3])) {
					status = atoi(bufp);
				}
				bufp = strchr(bufp, '\n')+1;
				continue;
			}
			else if (*bufp == 't' && ! strncmp(bufp, "total", 5)) {
				/* ^total -- total size in kilobytes */
				bufp = strchr(bufp, '\n')+1;
				continue;
			}
			else {
				dlog("FTP parse error: \"%32.32s...\"", bufp);
				return 550;
			}
		}

		/* /[\-rwxs]{9}/ -- permission */
		if (strchr("-rwxs", *bufp)) {
			for (i = 0; i < 9; ++i) {
				f_mode <<= 1;
				if (strchr("rwxs", *bufp))
					f_mode += 1;
				++bufp;
			}
			f_mode &= ~0222;
			if (f_mode & 0444) f_mode |= 0644;
			if (f_mode & 0111) f_mode |= 0111;
		} else {
			dlog("ftp_parse_list: unknown permission format");
			return 550;
		}

		while (isspace(*bufp)) ++bufp;
		/* /\d+/ -- nlinks */
		if (isdigit(*bufp)) {
			f_nlink = atoi(bufp);
			while (isdigit(*bufp)) ++bufp;
		} else {
			dlog("ftp_parse_list: no nlink");
			return 550;
		}

		while (isspace(*bufp)) ++bufp;
		/* /[^ ]+/ -- owner */
		while (! isspace(*bufp)) ++bufp;
		f_uid = 0;

		while (isspace(*bufp)) ++bufp;
		/* /[^ ]+/ -- group */
		while (! isspace(*bufp)) ++bufp;
		f_gid = 0;

		while (isspace(*bufp)) ++bufp;
		/* /\d+/ -- file size */
		if (isdigit(*bufp)) {
			f_size = atoi(bufp);
			while (isdigit(*bufp)) ++bufp;
		} else {
			dlog("ftp_parse_list: no size");
			return 550;
		}

		bzero(&tm, sizeof (struct tm));
		while (isspace(*bufp)) ++bufp;
		/* /[a-zA-Z]{3}/ -- month */
		if (! isalpha(*bufp)) {
			dlog("ftp_parse_list: no month");
			return 550;
		}
		switch (bufp[0] + bufp[1] + bufp[2]) {
		case 0x119:	/* Jan */
			tm.tm_mon = 0;
			break;
		case 0x10d:	/* Feb */
			tm.tm_mon = 1;
			break;
		case 0x120:	/* Mar */
			tm.tm_mon = 2;
			break;
		case 0x123:	/* Apr */
			tm.tm_mon = 3;
			break;
		case 0x127:	/* May */
			tm.tm_mon = 4;
			break;
		case 0x12d:	/* Jun */
			tm.tm_mon = 5;
			break;
		case 0x12b:	/* Jul */
			tm.tm_mon = 6;
			break;
		case 0x11d:	/* Aug */
			tm.tm_mon = 7;
			break;
		case 0x128:	/* Sep */
			tm.tm_mon = 8;
			break;
		case 0x126:	/* Oct */
			tm.tm_mon = 9;
			break;
		case 0x133:	/* Nov */
			tm.tm_mon = 10;
			break;
		case 0x10c:	/* Dec */
			tm.tm_mon = 11;
			break;
		default:
			dlog("ftp_parse_list: unknown month");
			return 550;
		}
		bufp += 3;

		while (isspace(*bufp)) ++bufp;
		/* /\d+/ -- day of month */
		tm.tm_mday = atoi(bufp);
		while (! isspace(*bufp)) ++bufp;

		while (isspace(*bufp)) ++bufp;
		/* /(\d\d:\d\d|\d{4})/ -- time or year */
		if (bufp[2] == ':') {
			tm.tm_year = cs_year;
			if (tm.tm_mon > cs_month) {
				--tm.tm_year;
			}
			tm.tm_hour = atoi(bufp);
			bufp += 3;
			tm.tm_min = atoi(bufp);
			bufp += 2;
		} else {
			tm.tm_year = atoi(bufp) - 1900;
			bufp += 4;
		}
		f_time = tm2time(&tm);

		while (isspace(*bufp)) ++bufp;
		/* /[^ ]+/ -- filename */
		if (f_type != NFLNK) {
			p = strchr(bufp, '\r');
			*p = '\0';
			if (*bufp == '.' && (strcmp(bufp, ".") == 0
					     || strcmp(bufp, "..") == 0)) {
				/*
				 * We do not put "." and ".." onto directory
				 * chain; update attributes where necessary.
				 */
				f = file_findname(dirp, bufp);
				assert(f != (wf_file *)0);
				f->fattr.nlink = f_nlink;
				f->fattr.atime.seconds
					= f->fattr.mtime.seconds
					= f->fattr.ctime.seconds = f_time;
				bufp = p + 2;		/* skip \r\n */
				continue;
			}
			f_name = strdup(bufp);
			if (f_type == NFDIR) {
				if (olddirp &&
				    (f = file_findname(olddirp, f_name))) {
					f_id = f->id;
				} else {
					f_id = vol_allocid(volp);
				}
			}
			*p = '\r';
		} else {
			p = bufp;
			while (p = strchr(p, '-')) {
				if (*++p == '>') break;
			}
			if (!p) {
				dlog("symlink botch");
				return 550;
			}
			*(p-2) = '\0';
			f_name = strdup(bufp);
			*(p-2) = ' ';
			bufp = p+2;
			p = strchr(bufp, '\r');	/* assuming \r\n */
			*p = '\0';
			f_link = strdup(bufp);
			*p = '\r';
		}
		f_fsid = volp->id;
		f = file_new(f_name, f_link, f_type, f_nlink, f_mode,
			 f_uid, f_gid, f_size, f_time, f_id, f_fsid);
		dir_addfile(dirp, f);
		bufp = strchr(bufp, '\n')+1;
	}
}

