/* 
 * WorldWide File System
 * Copyright (c) 1992,1993 Youki Kadobayashi
 * Copyright (c) 1992,1993 Osaka University
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that the following conditions
 * are met:
 * 1. Both the copyright notice and this permission notice appear in
 *    all copies of the software, derivative works or modified versions,
 *    and any portions thereof, and that both notices appear in
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Osaka University
 *      and its contributors.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE DEVELOPER ``AS IS'' AND OSAKA
 * UNIVERSITY DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
 * WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Osaka University requests users of this software to return to
 *
 *  Youki Kadobayashi
 *  Department of Information and Computer Sciences
 *  Osaka University, Toyonaka 560, Osaka, Japan
 *
 * any improvements or extensions that they make and grant Osaka
 * University the rights to redistribute these changes.
 */
/* nonblocking NFS and TCP */
static char *AtFSid = "$Header: nfs_start.c[109.0] Wed Nov 24 03:47:14 1993 youki-k@is.aist-nara.ac.jp saved $";

#include <sys/types.h>
#include <errno.h>
#include <setjmp.h>
#include <signal.h>
#include "wfs.h"
#include "util.h"
#include "global.h"

extern jmp_buf select_intr;
extern void nfs_program_2();
extern void cs_program_1();

SVCXPRT *nfsxprt;
SVCXPRT *csxprt;
fd_set	read_fdset, write_fdset, except_fdset;

#define MASKED_SIGS \
	(sigmask(SIGINT)|sigmask(SIGTERM)|sigmask(SIGHUP))

extern qelem so_head;
qelem so_head = { &so_head, &so_head };

char *sobuf[WF_MAX_SO];

typedef struct soq {
	qelem	q;
	int	so;
	func	funcp;
	void	*p;
} soq;

soq soq_ary[WF_MAX_SO];

void
so_register(so, flag)
int so;
int flag;
{
	if (flag & WF_SO_READ)
		FD_SET(so, &read_fdset);
	if (flag & WF_SO_WRITE)
		FD_SET(so, &write_fdset);
	if (flag & WF_SO_EXCEPT)
		FD_SET(so, &except_fdset);
}

void
so_unregister(so, flag)
int so;
int flag;
{
	if (flag & WF_SO_READ)
		FD_CLR(so, &read_fdset);
	if (flag & WF_SO_WRITE)
		FD_CLR(so, &write_fdset);
	if (flag & WF_SO_EXCEPT)
		FD_CLR(so, &except_fdset);
}

void
fd_union(destfdp, srcfdp)
fd_set *destfdp, *srcfdp;
{
	int i;
	for (i = 0; i < (sizeof(fd_set) / sizeof(fd_mask)); ++i) {
		destfdp->fds_bits[i] |= srcfdp->fds_bits[i];
	}
}

void
so_callback(so, funcp, p)
int so;
func funcp;
void *p;
{
	soq *sop = &soq_ary[so];
#ifndef PERFECTLY_CORRECT
	if (soq_ary[so].so) {
		dlog("so_callback: not cleared");
		return;
	}
#endif
	sop->so = so;
	sop->funcp = funcp;
	sop->p = p;
	q_insert(&sop->q, LAST(soq, &so_head));
}

void
so_clear(so)
int so;
{
	soq *sop = &soq_ary[so];
#ifndef PERFECTLY_CORRECT
	if (soq_ary[so].so == 0) {
		dlog("so_clear: already cleared");
		return;
	}
#endif
	soq_ary[so].so = 0;
	q_remove(&sop->q);
}

char *
so_getbuf(so)
int so;
{
	if (! sobuf[so]) {
		sobuf[so] = (char *) malloc(NFS_MAXDATA+1);
	}
	return sobuf[so];
}

static int
do_select(smask, fds, readfdp, writefdp, exceptfdp, tvp)
int smask, fds;
fd_set *readfdp, *writefdp, *exceptfdp;
struct timeval *tvp;
{
	int sig;
	int nsel;
	if (sig = setjmp(select_intr)) {
		select_intr_valid = 0;
		/* Got a signal */
		switch (sig) {
		case SIGINT:
		case SIGTERM:
			if (cs_state != Finishing) {
				cs_state = Finishing;
				/* attempt to terminate other threads */
				csd_stop();
			}
			break;
		}
		nsel = -1;
		errno = EINTR;
	} else {
		select_intr_valid = 1;
		/*
		 * Invalidate the current clock value
		 */
		now = 0;
		/*
		 * Allow interrupts.
		 * If a signal occurs, then it will cause longjump up above.
		 */
		(void) sigsetmask(smask);
		/*
		 * Wait for input
		 */
		nsel = select(fds, readfdp, writefdp, exceptfdp,
			      tvp->tv_sec ? tvp : (struct timeval *) 0);
	}
	(void) sigblock(MASKED_SIGS);		/* Disallow interrupts */
	return nsel;
}

serv_state
csd_run()
{
	fd_set readfds, writefds, exceptfds;
	struct timeval tvv;
	int nsel;
	int smask = sigblock(MASKED_SIGS);

	next_softclock = gettime();
	cs_state = Run;

	while ((int)cs_state <= (int)Finishing) {
		/*
		 * notify tasks
		 */
		if (task_notify_todo)
			do_task_notify();

		/*
		 * compute timeout tvv
		 */
		realtime();
		if (next_softclock <= now) {
			if (cs_state == Finishing) {
				/* XXX */;
			}
			tvv.tv_sec = softclock();
		} else {
			tvv.tv_sec = next_softclock - now;
		}
		tvv.tv_usec = 0;

		/*
		 * set fd masks
		 */
		bzero(&readfds, sizeof(readfds));
#ifdef RPC_4
		bcopy(&svc_fdset, &readfds, sizeof(svc_fdset));
#else
		bcopy(&svc_fds, &readfds, sizeof(svc_fds));
#endif
		fd_union(&readfds, &read_fdset);
		bcopy(&write_fdset, &writefds, sizeof(writefds));
		bcopy(&except_fdset, &exceptfds, sizeof(exceptfds));

		/*
		 * Wait for events to occur
		 */
		nsel = do_select(smask, WF_MAX_SO /* XXX */,
				 &readfds, &writefds, &exceptfds, &tvv);
		switch (nsel) {
		case -1:
			/*
			 * select was interrupted
			 */
			if (errno != EINTR) {
				dlog("select: errno = %d", errno);
#if 0
				dump_fdset("select readfds", &readfds);
				dump_fdset("select writefds", &writefds);
#endif
			}
			break;
		case 0:
			/*
			 * select timeout
			 */
			/* nothing */;
			break;
		default:
			if ((int)cs_state < (int)Finishing) {
				/*
				 * process queued socket jobs
				 */
				soq *sop, *sop2;
				ITER2(sop, sop2, soq, &so_head) {
					if (FD_ISSET(sop->so, &readfds)
					    || FD_ISSET(sop->so, &writefds)
					    || FD_ISSET(sop->so, &exceptfds)) {
						FD_CLR(sop->so, &readfds);
						(*sop->funcp)(sop->p);
						--nsel;
					}
				}
			}
			/*
			 * process RPC jobs
			 */
			if (nsel) {
#ifdef RPC_4
				svc_getreqset(&readfds);
#else
				svc_getreq(readfds.fds_bits[0]);
#endif /* RPC_4 */
			}
			break;
		}
		if (cs_state == Finishing && conn_count() == 0) {
			cs_state = Quit;
			break;
		}
	}
	(void) sigsetmask(smask);

	if (cs_state == Quit)
		cs_state = Done;
	return cs_state;
}

int
csd_rpc()
{
	int so;
	int ret;
	unsigned short port;

	/*
	 * create socket
	 */
	so = socket(AF_INET, SOCK_DGRAM, 0);
	if (so < 0) {
		perror("Can't create socket");
		return 1;
	}

	/*
	 * bind NFS port
	 */
	ret = bind_resv_port(so, &port);
	cs_nfsport = port;
	if (ret) {
		syslog(LOG_CRIT, "couldn't bind reserved port");
		return 1;
	}

	/*
	 * create RPC stub
	 */
	nfsxprt = svcudp_create(so);
	csxprt = svcudp_create(so);
	if (nfsxprt == NULL || csxprt == NULL) {
		syslog(LOG_CRIT, "cannot create rpc/udp service");
		return 2;
	}

	/*
	 * register NFS port
	 */
	if (! svc_register(nfsxprt, NFS_PROGRAM, NFS_VERSION,
			   nfs_program_2, 0)) {
		syslog(LOG_CRIT, "unable to register NFS_PROGRAM");
		return 3;
	}

	/*
	 * register CS port
	 */
	pmap_unset(CS_PROGRAM, CS_VERSION);
	if (! svc_register(csxprt, CS_PROGRAM, CS_VERSION,
			   cs_program_1, IPPROTO_UDP)) {
		syslog(LOG_CRIT, "unable to register CS_PROGRAM");
		return 3;
	}

	return 0;
}

