/* 
 * WorldWide File System
 * Copyright (c) 1992,1993 Youki Kadobayashi
 * Copyright (c) 1992,1993 Osaka University
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that the following conditions
 * are met:
 * 1. Both the copyright notice and this permission notice appear in
 *    all copies of the software, derivative works or modified versions,
 *    and any portions thereof, and that both notices appear in
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Osaka University
 *      and its contributors.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE DEVELOPER ``AS IS'' AND OSAKA
 * UNIVERSITY DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
 * WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Osaka University requests users of this software to return to
 *
 *  Youki Kadobayashi
 *  Department of Information and Computer Sciences
 *  Osaka University, Toyonaka 560, Osaka, Japan
 *
 * any improvements or extensions that they make and grant Osaka
 * University the rights to redistribute these changes.
 */
static char *AtFSid = "$Header: pol.c[109.0] Wed Nov 24 03:47:15 1993 youki-k@is.aist-nara.ac.jp saved $";

#include "wfs.h"
#include "util.h"
#include "global.h"

/*
 * Policy-based protocol selection. Work in progress.
 *
 * Appropriate protocol will be selected from protocol set, based on
 * the following four constraints:
 *
 * - Volume constraints. Protocols may be narrowed due to security concern,
 *   server load, or purpose of the volume.
 * - Session constraints (or server constraints). Protocol may be narrowed
 *   for better user response. This is effective particularly if there are
 *   existing connection to a server serving the volume.
 * - Cluster-site constraints. Site manager can specify certain protocol set
 *   to gain better security and/or performance.
 * - User constraints. User or user agent may prefer particular protcol set
 *   from others for some reason, such as provision of better semantics
 *   for the application in use.
 *
 * For more detailed discussion please expect my upcoming paper.
 * I haven't implemented this model so far because I don't need it right now.
 * It must be useful however, if you want to implement WAIS-FS or Archie-FS
 * or NNTP-FS or TCP-NFS...
 */
wf_proto *
pol_select_proto(volp)
wf_vol *volp;
{
	switch (volp->id) {
	case WF_ROOT_VOL_ID:
		return &proto_root;
#if 0
	case WF_HOST_VOL_ID:
		return &proto_host;
	case WF_CWD_VOL_ID:
		return &proto_cwd;
#endif
	case WF_FH_VOL_ID:
		return &proto_fh;
	}
	if (volp->flag & WF_VOL_DIRECTFILE) {
		return &proto_ufs;
	} else {
		return &proto_ftp;	/* XXX */
	}
}

/*
 * Policy-based server selection. Work in progress.
 *
 * A file server may be selected from given set of file servers, with
 * the following considerations for each server in choice:
 *
 * - Reliability of the candidate.
 * - Mean RTT between the candidate and the cluster server.
 * - Average load of the candidate.
 * - Policy; such as the number of connections the candidate permit.
 * - Protocol set the candidate can speak.
 */
wf_srvq *
pol_select_srv(volp)
wf_vol *volp;
{
	wf_srvq *q;
	ITER(q, wf_srvq, &volp->srvq->q) {
		if (q->srv->flag == 0)
			return q;
	}
	q = NEXT(wf_srvq, volp->srvq);	/* firewall */
	return q;
}
