/* 
 * WorldWide File System
 * Copyright (c) 1992,1993 Youki Kadobayashi
 * Copyright (c) 1992,1993 Osaka University
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that the following conditions
 * are met:
 * 1. Both the copyright notice and this permission notice appear in
 *    all copies of the software, derivative works or modified versions,
 *    and any portions thereof, and that both notices appear in
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Osaka University
 *      and its contributors.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE DEVELOPER ``AS IS'' AND OSAKA
 * UNIVERSITY DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
 * WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Osaka University requests users of this software to return to
 *
 *  Youki Kadobayashi
 *  Department of Information and Computer Sciences
 *  Osaka University, Toyonaka 560, Osaka, Japan
 *
 * any improvements or extensions that they make and grant Osaka
 * University the rights to redistribute these changes.
 */
static char *AtFSid = "$Header: trace.c[109.0] Wed Nov 24 03:47:18 1993 youki-k@is.aist-nara.ac.jp saved $";

#include <stdio.h>
#include "wfs.h"
#include "global.h"
#include "util.h"

extern qelem trace_head;
qelem trace_head = { &trace_head, &trace_head };	/* chain of wf_trace */
static void *trace_callout;

static int idle = 0;
static FILE *fp;
static int closed = 1;

void
trace_event(handle, tracep)
void *handle;
wf_trace *tracep;
{
	/* use casted pointer to discriminate sessions. */
	tracep->cookie = (long) handle;
	q_insert(tracep, LAST(wf_trace, &trace_head));
	idle = 0;
}

static void
trace_open()
{
	char path[MAXPATHLEN];

	sprintf(path, "%s/log/trace%s", cs_topdir, cs_date);
	fp = fopen(path, "a+");
	if (fp == NULL) {
		dlog("cannot write to trace file \"%s\"", path);
	} else {
		closed = 0;
	}
}

static void
trace_close()
{
	if (!closed) {
		fclose(fp);
		closed = 1;
	}
}

static void
trace_write()
{
#define TRACE_SIZE (sizeof (wf_trace) - sizeof (qelem))
	wf_trace *tracep;
	char *p, *buf;
	int count = 0;
	static int maxcount = NFS_MAXDATA / TRACE_SIZE;

	/*
	 * open file if necessary
	 */
	if (! MORE(&trace_head)) return;
	if (closed)
		trace_open();
	if (fp == NULL) return;

	/*
	 * write data
	 */
	p = buf = so_getbuf(fileno(fp));	/* of size NFS_MAXDATA */
	ITER(tracep, wf_trace, &trace_head) {
		bcopy(&tracep->cookie, p, TRACE_SIZE);
		p += TRACE_SIZE;
		++count;
		if (count >= maxcount) {
			fwrite(buf, TRACE_SIZE, count, fp);
			p = buf;
			count = 0;
		}
	}
	if (count) {
		fwrite(buf, TRACE_SIZE, count, fp);
	}

	/*
	 * forget data
	 */
	q_allfree(&trace_head);
}

/*
 * automagically close trace file if no activity has been observed recently.
 */
static void
trace_gc()
{
	trace_write();
	if (!closed) {
		idle += 60;
		if (idle > 120) {
			trace_close();
		}
	}
	trace_callout = timeout_set(gettime()+60, trace_gc, 0);
}

/* forcibly flush trace file and execute shell script. */
void
trace_report()
{
	char path[MAXPATHLEN];

	trace_write();
	trace_close();
	sprintf(path, "%s/bin/report.sh %s/log/trace%s %s.%s",
		cs_topdir,
		cs_topdir, cs_date, cs_hostname, cs_domain);
	unix_command(path);
}

void
trace_start()
{
	trace_callout = timeout_set(gettime()+60, trace_gc, 0);
}

