/* 
 * WorldWide File System
 * Copyright (c) 1992,1993 Youki Kadobayashi
 * Copyright (c) 1992,1993 Osaka University
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that the following conditions
 * are met:
 * 1. Both the copyright notice and this permission notice appear in
 *    all copies of the software, derivative works or modified versions,
 *    and any portions thereof, and that both notices appear in
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Osaka University
 *      and its contributors.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE DEVELOPER ``AS IS'' AND OSAKA
 * UNIVERSITY DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
 * WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Osaka University requests users of this software to return to
 *
 *  Youki Kadobayashi
 *  Department of Information and Computer Sciences
 *  Osaka University, Toyonaka 560, Osaka, Japan
 *
 * any improvements or extensions that they make and grant Osaka
 * University the rights to redistribute these changes.
 */
/* cached local filesystem */
static char *AtFSid = "$Header: ufs.c[109.0] Wed Nov 24 03:47:18 1993 youki-k@is.aist-nara.ac.jp saved $";

#include <sys/types.h>
#ifdef _POSIX_SOURCE
#include <dirent.h>
#define	DIRENT	struct dirent
#else
#include <sys/dir.h>
#define	DIRENT	struct direct
#endif

#include <sys/stat.h>
#include <fcntl.h>		/* for O_WRONLY etc. */
#include "wfs.h"
#include "util.h"
#include "global.h"

/* public functions  */
void ufs_read_miss(), ufs_readdir_miss();

wf_proto proto_ufs = {
	"UFS", cmfs_getattr, err_getattr_miss,
	cmfs_lookup, err_lookup_miss, cmfs_readlink,
	cmfs_read, ufs_read_miss, cmfs_readdir, ufs_readdir_miss,
	err_getuda, err_keepalive, err_close, err_shutdown, 0
};

void
ufs_read_miss(c)
wf_thrd *c;
/* in: rqstp, volp, offset, count, totalcount, dirp, filep */
/* out: */
{
	dlog("ufs_read_miss: should not be called...");
}

void
ufs_readdir_miss(c)
wf_thrd *c;
/* in: rqstp, volp, filep, fhp, cookie, count, dotdotp, olddirp */
/* out: */
{
	DIR *dp;
	struct direct *p;
	struct stat s;
	wf_file *f;
	ftype f_type;
	u_int f_id, namelen;
	char *f_name, *f_link, link[MAXPATHLEN];
	char path[MAXPATHLEN], path2[MAXPATHLEN];
	char *topdir = c->child_volp->topdir;

	/*
	 * create the missing directory first
	 */
	if (c->fhp->child_dir > WF_VOL_BASE) {
		if (c->olddirp) {
			c->dirp = dir_new(c->child_volp, c->fhp->child_dir,
					  c->dotdotp->name, "");
		} else {
			c->dirp = vol_newroot(c->child_volp);
		}
	} else {
		c->dirp = dir_new(c->child_volp, c->fhp->child_dir,
				  c->dotdotp->name, c->filep->name);
	}
	f = dir_adddir(c->dirp, ".", c->child_volp->id, c->dirp->id);
	attr_touch(&f->fattr);
	f = dir_adddir(c->dirp, "..", c->parent_volp->id, c->dotdotp->id);
	attr_touch(&f->fattr);
	/*
	 * read the corresponding UFS directory...
	 */
	sprintf(path, "%s%s", topdir, c->dirp->name);
	dp = opendir(path);
	if (! dp) {
		dlog("could not opendir \"%s\"", path);
		return;
	}
	while (p = readdir(dp)) {
		f_name = strdup(p->d_name);
		f_link = NULL;
		if (strcmp(f_name, ".") == 0 || strcmp(f_name, "..") == 0)
			continue;	/* ignore "." and ".." */
		sprintf(path2, "%s/%s", path, f_name);
		lstat(path2, &s);
		/*
		 * compute attributes for each cached directory entry
		 */
		f_id = 0;
		if (s.st_mode & S_IFREG) {
			f_type = NFREG;
		} else if (s.st_mode & S_IFLNK) {
			f_type = NFLNK;
			namelen = readlink(path2, link, MAXPATHLEN);
			if (namelen > 0) {
				link[namelen-1] = '\0';
				f_link = strdup(link);
			} else {
				/* could not stat it; return fake entry */
				f_link = strdup("/???");
			}
		} else if (s.st_mode & S_IFDIR) {
			f_type = NFDIR;
			if (c->olddirp &&
			    (f = file_findname(c->olddirp, f_name))) {
				f_id = f->id;
			} else {
				f_id = vol_allocid(c->child_volp);
			}
		}
		/*
		 * append it to the newly created directory
		 */
		f = file_new(f_name, f_link, f_type,
			 s.st_nlink, s.st_mode, s.st_uid, s.st_gid,
			 s.st_size, s.st_mtime, f_id, c->child_volp->id);
		dir_addfile(c->dirp, f);
	}
	closedir(dp);
	dir_save(c->child_volp, c->dirp, c->olddirp);
}
