/* 
 * WorldWide File System
 * Copyright (c) 1992,1993 Youki Kadobayashi
 * Copyright (c) 1992,1993 Osaka University
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that the following conditions
 * are met:
 * 1. Both the copyright notice and this permission notice appear in
 *    all copies of the software, derivative works or modified versions,
 *    and any portions thereof, and that both notices appear in
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Osaka University
 *      and its contributors.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE DEVELOPER ``AS IS'' AND OSAKA
 * UNIVERSITY DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
 * WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Osaka University requests users of this software to return to
 *
 *  Youki Kadobayashi
 *  Department of Information and Computer Sciences
 *  Osaka University, Toyonaka 560, Osaka, Japan
 *
 * any improvements or extensions that they make and grant Osaka
 * University the rights to redistribute these changes.
 */
/* $Header: util.h[109.1] Tue Dec  7 23:38:01 1993 youki-k@is.aist-nara.ac.jp saved $ */

/* Debugging malloc library of Conor P. Cahill */
#ifdef DEBUG_MALLOC
#include "/usr/local/include/dbmalloc/malloc.h"
#endif
/* If you want event traces */
/* #define TRACE */

#ifndef MIN
#define	MIN(a,b)	((a)<(b)?(a):(b))
#endif	/* !MIN */

#define WF_WRITE(fp, info)	fwrite(&(info), sizeof(info), 1, fp)
#define WF_READ(fp, info)	fread(&(info), sizeof(info), 1, fp)

#define WF_MAX_SO	64	/* machine dependent */
#define WF_SO_READ	0x1
#define WF_SO_WRITE	0x2
#define WF_SO_EXCEPT	0x4
#define WF_SO_ALL	0x7

#define WF_VOL_TTL	(30 * 60)	/* volume in-core lifetime */
#define WF_DIR_TTLMEM	(5 * 60)	/* directory in-core lifetime */
#define WF_DIR_TTLDISK	(24 * 60 * 60)
#define WF_CONN_TTL	(2 * 60)	/* idle connection lifetime */
#define WF_THREAD_TTL	(2 * 60)	/* sucked thead lifetime */

extern time_t	now;
#define gettime()	(now ? now : time(&now))
#define realtime()	(time(&now))

#ifdef DEBUG_MALLOC
#define	ALLOC(s)	(s *)alloc_and_clear(__FILE__, __LINE__, sizeof(s))
#define	CLONE(v)	alloc_and_copy(__FILE__, __LINE__, (v), sizeof(*(v)))
#define	db_alist	file, line, 
#define	db_dcl		char *file; int line;
#else
#define	ALLOC(s)	(s *)alloc_and_clear(sizeof(s))
#define	CLONE(v)	alloc_and_copy((v), sizeof(*(v)))
#define	debug_malloc(s)	malloc(s)
#define	db_alist
#define	db_dcl
#endif
#define	FREE(s)		free(s)

#define assert(e)	{ if (!(e)) {dlog("assertion failed at %s:%d", __FILE__, __LINE__); exit(1); } }
#define	warn_negative(a)	{ if ((a) < 0) {dlog("should not take negative value at %s:%d", __FILE__, __LINE__); }}
#define	warn_zero(a)	{ if ((a) < 0) {dlog("should not compute to zero at %s:%d", __FILE__, __LINE__); }}
#define	PROT(s)		((s) ? (s) : "(null)")

#define errno_diag()	errno_log(__FILE__, __LINE__)

#include <sys/syslog.h>
