#!/usr/local/bin/perl

# DNS resolver library.  See RFC1035 for more details.
# by Marc Horowitz <marc@mit.edu>

# $Id: resolv.pl,v 1.3 91/11/09 00:35:49 marc Exp $

# The interface works like this (all functions are in the main package)
#
# res_init()
#	One-time initialization
#
# res_open($nameserver)
#	Sets up a connection to $nameserver.
#	If $nameserver is undefined or empty, then the server is
#	looked up in /etc/resolv.conf.  Returns the filehandle.
#
# res_mkquery($name, $type, $class, $id)
#	builds and returns a query for the given name, type, class,
#	and id.  If no id is given, a random ID is chosen.
#
# res_send($fh, $query)
#	sends $query to the server at $fh.  Returns the response from the
#	server.
#
# response format:
# [0] = id
# [1] = authoritative
# [2] = recursion available
# [3] = query name
# [4] = query class
# [5] = query type
# [6] = start of answers
# [7] = end of answers
# [8] = start of authority records
# [9] = end of authority records
# [10] = start of add'l records
# [11] = end of add'l records
# [12] ...   resource records (in multiples of 5)
#
# The variable res'options can be set to any of the values
# which _res.options can be set to.  The constants will be in the 
# main package.
#
# Not implemented yet:
# Only one query is sent out.  That is, RES_DNSRCH is ignored, and
#	only the primary nameserver is used.
# RES_DEBUG doesn't print anything
#	

## Everything after this is code.  If you look at it, you're
## violating an abstraction barrier.  Shame on you.  :-)

# hack! hack!  This is to confuse the byte order stuff in arpa/nameser.h
# nothing here depends on it anyway.

sub MIPSEB {1;}

require 'sys/socket.ph' || die "can\'t do sys/socket.ph: $@";
require 'arpa/nameser.ph' || die "can\'t do arpa/nameser.ph: $@";
require 'resolv.ph' || die "can\'t do resolv.ph: $@";

# who? me? kludge?

undef &MIPSEB;

package res;

# Create conversion arrays unfortunately, I need to hardcode lists
# of the types and classes.

@qtypes = ("A", "NS", "MD", "MF", "CNAME", "SOA", "MB", "MG", "MR", "NULL",
	    "WKS", "PTR", "HINFO", "MINFO", "MX", "TXT", "UINFO", "UID", "GID",
	    "UNSPEC", "UNSPECA", "AXFR", "MAILB", "MAILA", "ANY");

@qclasses = ("IN", "CHAOS", "HS", "ANY");

for (@qtypes) {
    eval "\@qtype[&main'T_$_] = \"$_\";\$qtype{\$_} = &main'T_$_;";
}

for (@qclasses) {
    eval "\@qclass[&main'C_$_] = \"$_\";\$qclass{\$_} = &main'C_$_;";
}

sub qtype_strtonum {
    return($qtype{$_[0]} || -1);
}

sub qclass_strtonum {
    return($qclass{$_[0]} || -1);
}

sub qtype_numtostr {
	local($str) = @qtype[$_[0]];

	if (defined($str)) {
		return($str);
	} else {
		return("$_[0]");
	}
}

sub qclass_numtostr {
	local($str) = @qclass[$_[0]];

	if (defined($str)) {
		return($str);
	} else {
		return("$_[0]");
	}
}

# option bits and defaults

$options = &main'RES_DEFAULT;
$domain = "";

sub debug { $options & &main'RES_DEBUG; }

sub ipaddr {
    if ($_[0] =~ /(\d+)\.(\d+)\.(\d+)\.(\d+)/) {
	$saddr = pack("CCCC", $1, $2, $3, $4);
    } else {
	($saddr = ((gethostbyname($_[0]))[4] || undef)) ||
	    return(undef);
    }
}

sub read_conf {
    return if ($domain);

    local($addr);

    print "Reading /etc/resolv.conf\n" if &debug;

    if (open(CONF,"/etc/resolv.conf")) {
	local($") = ",";
	while(<CONF>) {
	    chop;
	    if (/^domain\s+/) { $domain = $'; }
	    elsif (/^nameserver\s+/) {
		if ($addr = &ipaddr($')) { push(@servers,$addr); }
	    }
	}
	close(CONF);
    } else {
	$domain = '';
	@servers = ("\0\0\0\0");
    }
    print "domain is ",$domain,"\nresolvers are @servers\n" if &debug;
}

# one-time initialization

$inited = 0;

sub main'res_init {
    return if ($options & &main'RES_INIT);

    local($fh);

    socket(NSUDP, &main'AF_INET, &main'SOCK_DGRAM, &main'PF_UNSPEC) ||
	die "socket: $!";

    $options |= &main'RES_INIT;
    &read_conf;

    $fh = select(NSUDP); $| = 1;   # set nonbufferred
    select($fh);
}

# begin a conversation with a nameserver

sub main'res_open {			# @_ = ($nameserver)
    local($saddr,$port);

    &main'res_init if (!($options & &main'RES_INIT));

    if ($_[0]) {
	if ($_[0] =~ /(\d+)\.(\d+)\.(\d+)\.(\d+)/) {
	    $saddr = pack("CCCC", $1, $2, $3, $4);
	} else {
	    ($saddr = ((gethostbyname($_[0]))[4] || undef)) ||
		return(undef);
	}
    } else {
	$saddr = $servers[0];
    }

    # lossage in /etc/services.  hardcode for now.
    #$port = (getservbyname("nameserver","tcp"))[2];
    $port = 53;

    $port = pack("S n a4 x8",&main'AF_INET,$port,$saddr);

    return($port);
}

sub main'res_mkquery { # @_ = ($name,$type,$class,$id)
    local($question);

    print "building query for <$_[0],$_[1],$_[2]>\n" if &debug;

    $question = (($options & &main'RES_RECURSE)? #  header
		 pack("n6",$_[3],0x0100,1,0,0,0):
		 pack("n6",$_[3],0x0000,1,0,0,0));
    $question .= ((($options & &main'RES_DEFNAMES) &&
		   (substr($_[0],-1,1) ne "."))?
		  &unparse_name($_[0].".".$domain):
		  &unparse_name($_[0]));
    $question .= pack("n n",&qtype_strtonum($_[1]), # query
		      &qclass_strtonum($_[2]));
}

sub unparse_name {
	local($label,$labellen,$str);
	$str = "";
	foreach $label (split(/\./,$_[0])) {
		$labellen = length($label);
		$str .= pack("Ca$labellen",$labellen,$label);
	}
	$str .= pack("C",0);		# root octet
}

%nstcp = ();			# assoc array of open VC's
$sockcnt = "nstcp000";

sub main'res_send { # @_ = ($socket,$query)
    local($sin,$len,$packet) = ($_[0],pack("n",length($_[1])),$_[1]);
    local($fh,$resp);

    if ($options & &main'RES_USEVC) { # 
	$fh = $nstcp{$sin};
	if (!$fh) {
	    $fh = $sockcnt++;

	    if (&debug) {
		local($,) = (".");
		print "Connecting to ",(unpack("S n C4",$sin))[2..5,1],"\n";
	    }

	    socket($fh,&main'AF_INET,&main'SOCK_STREAM,&main'PF_UNSPEC) ||
		die "socket: $!";
	    connect($fh, $sin) || die "connect: $!";

	    $nstcp{$sin} = $fh;
	}

	if (&debug) {
	    local($,) = (".");
	    print "Sending to ",(unpack("S n C4",$sin))[2..5,1]," via TCP\n";
	}

	send($fh,$len,0);
	send($fh,$packet,0);

	if (&debug) {
	    local($,) = (".");
	    print "Receiving from ",(unpack("S n C4",$sin))[2..5,1]
		," via TCP\n";
	}

	read($fh,$len,2) || die "recv: $!,$@";
	if ($len = unpack("n",$len)) {
	    read($fh,$resp,$len) || die "recv: $!";
	}

	print "Got a response (size=$len)\n" if (&debug);

	if (! ($options & &main'RES_STAYOPEN)) {
	    foreach (keys(%nstcp)) {
		if (&debug) {
		    local($,) = (".");
		    print "Disconnecting from ",(unpack("S n C4",$_))[2..5,1]
			,"\n";
		}

		close($_);
	    }
	    %nstcp = ();
	}
    } else {
	if (&debug) {
	    local($,) = (".");
	    print "Sending to ",(unpack("S n C4",$sin))[2..5,1]," via UDP\n";
	}

	send(NSUDP,$packet,0,$sin) || die "send: $!";

	if (&debug) {
	    local($,) = (".");
	    print "Receiving from ",(unpack("S n C4",$sin))[2..5,1]," via UDP\n";
	}

	$len = 512;
	read(NSUDP,$resp,512,0) || die "recv: $!";

	print "Got a response (size=$len)\n" if (&debug);
    }

    &parse_response($resp);
}

sub parse_response { # @_ = ($response)
	local($response,@resp,@ptr) = @_;

	@ptr = ($response,0);

	$header = &next_chars(12,@ptr);
	($id,$bits,$qdcount,$ancount,$nscount,$adcount) = unpack("n6",$header);
	$auth = ($bits >> 10) & 0x01;
	$recurse = ($bits >> 8) & 0x01;
	if ((!$auth) && ($options & &main'RES_AAONLY)) {
	    ($ancount,$nscount,$adcount) = (0,0,0);
	}
	    
	$rrs = $ancount+$nscount+$adcount;
	@resp = ($id,$auth,$recurse);		# 		[0..2]

	push(@resp,&parse_name(@ptr)); 		# QNAME		[3]
	push(@resp,&next_netshort(@ptr));	# QTYPE		[4]
	push(@resp,&next_netshort(@ptr));	# QCLASS	[5]

	push(@resp,12);				# 		[6]
	push(@resp,@resp[$#resp]+5*$ancount-1);	#		[7]

	push(@resp,@resp[$#resp]+1);		#		[8]
	push(@resp,@resp[$#resp]+5*$nscount-1);	#		[9]

	push(@resp,@resp[$#resp]+1);		#		[10]
	push(@resp,@resp[$#resp]+5*$adcount-1);	#		[11]

	for ($i = 0 ; $i < $rrs ; $i++) {
	    push(@resp,&parse_rrbits(@ptr));
	}
	return(@resp);
}

sub parse_name {
	local($name,$ch,$ptr,@temp) = ("",substr($_[0],$_[1],1));
	while (ord($ch = substr($_[0],$_[1],1)) != 0) {
		# Message compression (RFC1035 4.1.4)
		if (ord($ch) >= 0xc0) {
			$ptr = &next_netshort(@_) & 0x3fff;
			@temp=($_[0],$ptr);
			$name .= "".&parse_name(@temp);
			return($name);
		}
		$name .= &next_str(@_).".";
	}
	&next_chars(1,@_);	# move past \0
	if ($name eq "") { $name = ".."; }
	chop($name);  # remove trailing "."
	return($name);
}

sub parse_rrbits {
	local(@rrec,$name,$rdlen,$pfct);
	@rrec = ();

	$name = &parse_name(@_);
	# if NAME is an odd number of bytes, eat an extra byte
	if (($name == "") || (length($name)%2 == 1)) {&next_chars(1,$_[0]);}
	@rrec = ($name);					# NAME
	push(@rrec,&qtype_numtostr(&next_netshort(@_)));	# TYPE
	push(@rrec,&qclass_numtostr(&next_netshort(@_)));	# CLASS
	push(@rrec,&next_netlong(@_));				# TTL (integer)

	$rdlen = &next_netshort(@_);
#	push(@rrec,&next_chars($rdlen,@_));			# RDATA
	$pfct = "rrparse_".&qtype_numtostr($rrec[1]);
	push(@rrec,&$pfct(@_,$rdlen));				# RDATA
	return(@rrec);
}

sub next_netshort {unpack("n",&next_chars(2,@_));}
sub next_netlong  {unpack("N",&next_chars(4,@_));}

# strips the first character-string from the argument, and returns it as a
# perl string
sub next_str {
	local($cslen);
	$cslen = unpack("C",&next_chars(1,@_));
	&next_chars($cslen,@_);
}

# returns the first $_[0] chars at position $_[2] in string $_[1]
# and increments $_[2]
sub next_chars {
	local($len,$str) = (length($_[1]),$_[1]);
	$_[2] += $_[0];
	substr($_[1],$_[2]-$_[0],$_[0]);
}

# routines to parse apart the rrdata

sub rrparse_A {
    join('.',unpack("C4",&next_chars(4,@_)));
}

sub rrparse_NS {
    join('.',&parse_name(@_));
}

sub rrparse_MD {
    join('.',&parse_name(@_));
}

sub rrparse_MF {
    join('.',&parse_name(@_));
}

sub rrparse_CNAME {
    join('.',&parse_name(@_));
}

sub rrparse_SOA {
    join(' ',
	 join('.',&parse_name(@_)),
	 join('.',&parse_name(@_)),
	 &next_netlong(@_),
	 &next_netlong(@_),
	 &next_netlong(@_),
	 &next_netlong(@_),
	 &next_netlong(@_));
}

sub rrparse_MB {
    join('.',&parse_name(@_));
}

sub rrparse_MG {
    join('.',&parse_name(@_));
}

sub rrparse_MR {
    join('.',&parse_name(@_));
}

sub rrparse_NULL {
    return(&next_chars($_[2],@_));
}

sub rrparse_WKS {
    local(@resp,$bitmap,$len,%bit);

    push(@resp,
	 &rrparse_A(@_),
	 ord(&next_chars(1,@_)));
    $bitmap = &next_chars($_[2]-5,@_);
    @bit{0..(length($bitmap)*8)} = split(//,unpack("B*", $bitmap));
    push(@resp, grep($bit{$_},keys(%bit)));

#    for (unpack("B$len", $bitmap)) {
#	push(@resp, $_) if $_;
#    }

#    for ($i=0;$i<length($bitmap)*8;$i++) {
#	if (vec($bitmap,$i,1)) { push(@resp,$i); }
#    }
    join(' ',@resp);
}

sub rrparse_PTR {
    join('.',&parse_name(@_));
}

sub rrparse_HINFO {
    &next_str(@_).$;.&next_str(@_);
}

sub rrparse_MINFO {
    join('.',&parse_name(@_)).",".join('.',&parse_name(@_));
}

sub rrparse_MX {
    &next_netshort(@_)." ".join('.',&parse_name(@_));
}

sub rrparse_TXT {
    local(@data,$end);
    $end = $_[1]+$_[2];
    while ($_[1] < $end) {
	push(@data,&next_str(@ptr));
    }
    if ($_[1] > $end) { print STDERR "Something weird in rrparse_TXT\n"; }
    join($;,@data);
}

sub rrparse_UNSPEC {
    &next_chars($_[2],@_);
}

sub rrparse_UNSPECA {
    &next_chars($_[2],@_);
}

sub rrparse {
    local($type,$class) = (&qtype_numtostr($_[1]),&qclass_numtostr($_[2]));
    local(@rrout);

    eval("\@rrout = &rrparse_$type(\$_[4])") || (@rrout = ("< ".$_[4]." >"));
    return($_[0],$type,$class,$_[3],@rrout);
}

sub qdparse {
    return($_[0],&qtype_numtostr($_[1]),&qclass_numtostr($_[2]));
}
