/* 
 * WorldWide File System
 * Copyright (c) 1992,1993 Youki Kadobayashi
 * Copyright (c) 1992,1993 Osaka University
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that the following conditions
 * are met:
 * 1. Both the copyright notice and this permission notice appear in
 *    all copies of the software, derivative works or modified versions,
 *    and any portions thereof, and that both notices appear in
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Osaka University
 *      and its contributors.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE DEVELOPER ``AS IS'' AND OSAKA
 * UNIVERSITY DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
 * WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Osaka University requests users of this software to return to
 *
 *  Youki Kadobayashi
 *  Department of Information and Computer Sciences
 *  Osaka University, Toyonaka 560, Osaka, Japan
 *
 * any improvements or extensions that they make and grant Osaka
 * University the rights to redistribute these changes.
 */
/*
 * libww: WWFS Application Programming Interface.
 * This file is part of WWFS, the WorldWide File System.
 *
 * You do not want to link your program with libww.a if you have spliced
 * libc routines in libc_wwfs.a and you are planning to link with it.
 */
static char *AtFSid = "$Header: libww.c[109.0] Wed Nov 24 03:49:15 1993 youki-k@is.aist-nara.ac.jp saved $";

#include <sys/errno.h>
extern int errno;
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
extern  int h_errno;	/* for BSDs */

#ifdef _POSIX_SOURCE
#include <dirent.h>
#define	DIRENT	struct dirent
#else
#include <sys/dir.h>
#define	DIRENT	struct direct
#endif
#include <fcntl.h>
#include <stdio.h>
#if defined(__STDC__) || defined(STDC_HEADERS)
#include <stdlib.h>
#endif
#include <string.h>
#define	LIBWW
#include "libww.h"
#include "parse.h"

#define	UIP_TCP_PORT	8002
#define	UIP_UDP_PORT	8002

#if 0
static char	cs_hostname[MAXHOSTNAMELEN];
static struct sockaddr_in cs_sin;
#endif
static char	cs_path[MAXPATHLEN];
static char	*ww_mountdir;
static struct sockaddr_in *ww_cshost;
static int
	uip_block=1,		/* block by default */
	uip_progress=0;		/* don't show percentage indicator */

#if 0
char *
ww_getaddr(path)
char *path;
{
	/*
	 * Get the CS service address.  We need pathname here, since
	 * clients can mount several CSs at different mount points.
	 * Gross hack is deployed here because we never want to deal
	 * with the diverse format of /etc/mtab.  NFS protocol should
	 * have provided "getnfsinfo" operation that provides rudimentary
	 * information, so that we could contact the machine serving
	 * the filesystem via other network services...
	 */

	char *p;
	struct stat statbuf;
	int fd;
	struct hostent *hp;
	int retry = 0;

	/* "/mount/point/volume/dir/.wwfs_host" */
	strcpy(cs_path, path);
	p = strrchr(cs_path, '/');
	if (p)
		strcpy(p+1, ".wwfs_host");
	else
		strcpy(cs_path, ".wwfs_host");	/* within "." */

	/* get the hostname of CS */
	fd = Open(cs_path, O_RDONLY, 0400);
	if (fd < 0)
		return NULL;		/* not served by CS */
	fstat(fd, &statbuf);
	read(fd, cs_hostname, statbuf.st_size);
	close(fd);

 dns_retry:
	hp = gethostbyname(cs_hostname);
	if (hp == (struct hostent *) 0) {
		if (h_errno == TRY_AGAIN) {
			if (++retry < 5) {
				sleep(1);
				goto dns_retry;
			}
		}
		return NULL;
	}
	cs_sin.sin_family = hp->h_addrtype;
	/*
	 * I use primary address here. If your cluster server have
	 * more than one interfaces, and if ipforwarding was set to 0,
	 * and if your client cannot reach primary address due to
	 * dropped ipforwarding, your network configuration is wrong.
	 */
	bcopy(hp->h_addr_list[0], &cs_sin.sin_addr, hp->h_length);
	return cs_hostname;
}

char *
ww_path(path)
char *path;
{
	/*
	 * Convert client-specific NFS pathname into CS's pathname,
	 * so that clients and CS can share a common namespace.
	 * e.g., "/mount/point/volume/dir/filename" -> "/volume/dir/filename"
	 */

	char *p;
	struct stat statbuf;
	int fd;

	/* "/mount/point/volume/dir/.wwfs_cwd" */
	strcpy(cs_path, path);
	p = strrchr(cs_path, '/');
	if (p)
		strcpy(p+1, ".wwfs_cwd");
	else
		strcpy(cs_path, ".wwfs_cwd");	/* within "." */

	/* get the current working directory within CS's namespace */
	fd = Open(cs_path, O_RDONLY, 0400);
	if (fd < 0)
		return NULL;		/* not served by CS */
	fstat(fd, &statbuf);
	read(fd, cs_path, statbuf.st_size);
	close(fd);

	/* Finally we get it: "/volume/dir/filename" */
	p = strrchr(path, '/');		/* basename */
	if (p)
		strcat(cs_path, p+1);
	else
		strcat(cs_path, path);
	return cs_path;
}
#endif

char *
ww_path(path)
char *path;
{
	/*
	 * Convert client-specific NFS pathname into CS's pathname,
	 * so that clients and CS can share a common namespace.
	 * e.g., "/mount/point/volume/dir/filename" -> "+filehandle/filename"
	 */

	char *p;
	struct stat statbuf;
	int fd;

	/* "/mount/point/volume/dir/.wwfs_fh" */
	strcpy(cs_path, path);
	p = strrchr(cs_path, '/');
	if (p)
		strcpy(p+1, ".wwfs_fh");
	else
		strcpy(cs_path, ".wwfs_fh");	/* within "." */

	/* get the filehandle */
	fd = Open(cs_path, O_RDONLY, 0400);
	if (fd < 0)
		return NULL;		/* not served by csd */
	fstat(fd, &statbuf);
	cs_path[0] = '+';
	read(fd, cs_path+1, statbuf.st_size);
	close(fd);

	strcat(cs_path, "/");
	/* Finally we get it: "+filehandle/filename" */
	p = strrchr(path, '/');		/* basename */
	if (p)
		strcat(cs_path, p+1);
	else
		strcat(cs_path, path);
	return cs_path;
}

static int
ww_set_cshost(cshost)
char *cshost;
{
	struct hostent *hp;
	int retry = 0;

	if (getenv("WW_CSHOST")) {
		/* You may want to use another csd, e.g., for debugging... */
		cshost = getenv("WW_CSHOST");
	}
	ww_cshost = malloc(sizeof (struct sockaddr_in));

 dns_retry:
	hp = gethostbyname(cshost);
	if (hp == (struct hostent *) 0) {
		if (h_errno == TRY_AGAIN) {
			if (++retry < 5) {
				sleep(1);
				goto dns_retry;
			}
		}
		return NULL;
	}
	ww_cshost->sin_family = hp->h_addrtype;
	/*
	 * I use primary address here. If your cluster server have
	 * more than one interfaces, and if ipforwarding was set to 0,
	 * and if your client cannot reach primary address due to
	 * dropped ipforwarding, your network configuration is wrong.
	 */
	bcopy(hp->h_addr_list[0], &ww_cshost->sin_addr, hp->h_length);
	return 0;
}

static int
ww_set_mountdir(mountdir)
char *mountdir;
{
	if (getenv("WW_MOUNTDIR")) {
		/* You may want to use another csd, e.g., for debugging... */
		int len;
		char *p;

		mountdir = getenv("WW_MOUNTDIR");
		len = strlen(mountdir)+1;
		p = malloc(len);	/* GonzOS lacks strdup() */
		bcopy(mountdir, p, len);
	}
	ww_mountdir = mountdir;
	return 0;
}

static int
ww_set_dummy(dummy)
char *dummy;
{
	return 0;
}

static wf_tailor ww_tailor[] = {
	{ "cshost", ww_set_cshost },
	{ "mountcmd", ww_set_dummy },
	{ "mountdir", ww_set_mountdir },
	{ NULL, (int (*)())0 }
};

static int
ww_bind(name, value)
char *name;
char *value;
{
	return parse_tailor(ww_tailor, name, value);
}

char *
ww_localpath(path)
char *path;
{
	/*
	 * Convert CS's pathname into client-specific NFS pathname,
	 * so that clients can access it with canonical pathname.
	 * e.g., "/volume/dir/filename" -> "/mount/point/volume/dir/filename"
	 */

	if (! ww_mountdir) {
		if (parse_conf("/etc/wwfs.conf", ww_bind) < 0)
			return NULL;	/* syntax error, or not a csd client */
		if (! ww_mountdir)
			return NULL;
	}
	strcpy(cs_path, ww_mountdir);
	strcat(cs_path, path);
	return cs_path;
}

int
ww_control(optname, optval)
int optname;
void *optval;
{
	/* Control the behavior of wwfs client process. */
	switch (optname) {
	case UIP_BLOCK:
		uip_block = *(int *)optval;
		break;
	case UIP_PROGRESS:
		uip_progress = *(int *)optval;
		break;
	default:
		return -1;
	}
	return 0;
}

#define	BUFLEN	1400		/* Ethernet MTU might be a good measure */
static char buf[BUFLEN];

static int
do_socket()
{
	int so, ret;
	struct sockaddr_in sin;

	so = socket(AF_INET, SOCK_DGRAM, 0);
	bzero(&sin, sizeof(sin));
	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = htonl(INADDR_ANY);
	ret = bind(so, (struct sockaddr *) &sin, sizeof (sin));
	if (ret < 0) {
		close(so);
		return -1;
	}
	return so;
}

static void
do_sendto(so, port, buf)
int so, port;
char *buf;
{
	ww_cshost->sin_port = htons(port);
	sendto(so, buf, strlen(buf), 0,
	       (struct sockaddr *) ww_cshost, sizeof(*ww_cshost));
}

static void
do_recvfrom(so, buf)
int so;
char *buf;
{
	struct sockaddr_in from;
	int fromlen;

 retry:
	fromlen = sizeof(from);
	recvfrom(so, buf, BUFLEN, 0,
		 (struct sockaddr *)&from, &fromlen);
	if (from.sin_addr.s_addr != ww_cshost->sin_addr.s_addr)
		goto retry;
	return;
}

int
ww_get(subpath)
char *subpath;
{
	/*
	 * No libc counterpart -- just retrieve a file/directory
	 * represented by a WWFS pathname (i.e., pathname with
	 * leading mount point dropped).
	 */
	int so;

	if (! ww_cshost) {
		if (parse_conf("/etc/wwfs.conf", ww_bind) < 0) {
			errno = EINVAL;	/* invalid configuration */
			return -1;	/* syntax error, or not a csd client */
		}
		if (! ww_cshost) {
			errno = EINVAL;	/* invalid configuration */
			return -1;
		}
	}
	if (uip_block) {
		sprintf(buf, "RETR %s\n", subpath);
		so = do_socket();
		if (so < 0)
			return -1;
		do_sendto(so, UIP_UDP_PORT, buf);
		while (1) {
			do_recvfrom(so, buf);
			/* parse it */
			if (buf[3] == '.') {
				if (buf[0] == '2') {
					close(so);
					return 0;
				}
				else if (buf[0] == '5') {
					errno = EIO;
					close(so);
					return -1;
				}
			}
		}
	} else if (uip_progress) {
	}
	return 0;
}

int
ww_nfsget(path)
char *path;
{
	/*
	 * No libc counterpart -- just retrieve a file/directory
	 * represented by an NFS pathname.
	 */
	char *subpath;
	int ret;

	if (subpath = ww_path(path)) {
		ret = ww_get(subpath);
		if (ret < 0)
			return ret;
	}
}

int
ww_open(path, flags, mode)
char *path;
int flags;
int mode;
{
	/* Open the file, with pre- and post-processing as specified by
	 * ww_control().
	 */
	char *subpath;
	int ret;

	if (subpath = ww_path(path)) {
		ret = ww_get(subpath);
		if (ret < 0)
			return ret;
		/* possibly other protocols */
	}
	return Open(path, flags, mode);
}

FILE *
ww_fopen(path, flags)
char *path;
char *flags;
{
	/* FOpen the file, with pre- and post-processing as specified by
	 * ww_control().
	 */
	char *subpath;
	int ret;

	if (subpath = ww_path(path)) {
		ret = ww_get(subpath);
		if (ret < 0)
			return NULL;
		/* possibly other protocols */
	}
	return Fopen(path, flags);
}

DIR *
ww_opendir(path)
char *path;
{
	/* Open the dir, with pre- and post-processing as specified by
	 * ww_control().
	 */
	char *subpath;
	int ret;

	if (subpath = ww_path(path)) {
		ret = ww_get(subpath);
		if (ret < 0)
			return NULL;
		/* possibly other protocols */
	}
	return Opendir(path);
}

