/* 
 * WorldWide File System
 * Copyright (c) 1992,1993 Youki Kadobayashi
 * Copyright (c) 1992,1993 Osaka University
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that the following conditions
 * are met:
 * 1. Both the copyright notice and this permission notice appear in
 *    all copies of the software, derivative works or modified versions,
 *    and any portions thereof, and that both notices appear in
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Osaka University
 *      and its contributors.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE DEVELOPER ``AS IS'' AND OSAKA
 * UNIVERSITY DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
 * WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Osaka University requests users of this software to return to
 *
 *  Youki Kadobayashi
 *  Department of Information and Computer Sciences
 *  Osaka University, Toyonaka 560, Osaka, Japan
 *
 * any improvements or extensions that they make and grant Osaka
 * University the rights to redistribute these changes.
 */
/* RPC stubs for CS (WWFS Cluster Server) protocol. */
/* $Header: cs_prot.x[109.0] Wed Nov 24 03:50:23 1993 youki-k@is.aist-nara.ac.jp saved $ */

const CS_MAXDATA	= 8192;		/* as per NFS */
const CS_FHSIZE		= 32;		/* as per NFS */
const CS_MAXPATHLEN	= 1024;		/* sys/param.h */
const CS_MAXNAMLEN	= 255;		/* sys/dir.h etc., vary */
const CS_MAXHOSTLEN	= 64;		/* sys/param.h */

enum cs_stat {
	CS_OK = 0,		/* Okay */
	CS_BULK = 201,		/* Bulk data transfer */
	CS_WAIT = 202,		/* Operation would block */
	CSERR_NOENT = 2,	/* No such file or directory */
	CSERR_ACCES = 13,	/* Access denied */
	CSERR_NOTDIR = 20,	/* Not a directory */
	CSERR_ISDIR = 21,	/* Is a directory */
	CSERR_FBIG = 27,	/* File too large */
	CSERR_NOSPC = 28,	/* No space left on device */
	CSERR_ROFS = 30,	/* Read-only file system */
	CSERR_STALE = 70,	/* Stale filehandle */
	CSERR_OTHER = 100,	/* Other errors; use geterr */
	CSERR_ALREADY = 101,	/* Operation already in progress */
	CSERR_PROTO = 102,	/* Protocol not available */
	CSERR_NDOWN = 103,	/* Network is down */
	CSERR_NUNREACH = 104,	/* Network unreachable */
	CSERR_XABORT = 105,	/* Transfer aborted */
	CSERR_TIMEOUT = 106,	/* Connection timed out */
	CSERR_REFUSED = 107,	/* Connection refused */
	CSERR_HDOWN = 108,	/* Host is down */
	CSERR_HUNREACH = 109,	/* No route to host */
	CSERR_NQUOT = 110	/* Network quota exceeded */
};

#ifndef RPC_XDR
%#ifndef NFS_FH_DEFINED

/* NFS compatible type definitions */
struct nfs_fh {
	opaque data[CS_FHSIZE];
};

typedef string filename<CS_MAXNAMLEN>; 
typedef string nfspath<CS_MAXPATHLEN>;

struct nfstime {
	unsigned seconds;
	unsigned useconds;
};

enum ftype {
	NFNON = 0,	/* non-file */
	NFREG = 1,	/* regular file */
	NFDIR = 2,	/* directory */
	NFBLK = 3,	/* block special */
	NFCHR = 4,	/* character special */
	NFLNK = 5,	/* symbolic link */
	NFSOCK = 6,	/* unix domain sockets */
	NFBAD = 7,	/* unused */
	NFFIFO = 8 	/* named pipe */
};

struct fattr {
	ftype type;		/* file type */
	unsigned mode;		/* protection mode bits */
	unsigned nlink;		/* # hard links */
	unsigned uid;		/* owner user id */
	unsigned gid;		/* owner group id */
	unsigned size;		/* file size in bytes */
	unsigned blocksize;	/* prefered block size */
	unsigned rdev;		/* special device # */
	unsigned blocks;	/* Kb of disk used by file */
	unsigned fsid;		/* device # */
	unsigned fileid;	/* inode # */
	nfstime	atime;		/* time of last access */
	nfstime	mtime;		/* time of last modification */
	nfstime	ctime;		/* time of last change */
};
/* end of NFS compatible type definitions */

%#define NFS_FH_DEFINED
%#endif
#endif

union cs_fhres switch (cs_stat status) {
	case CS_OK:
		nfs_fh	file;
	default:
		long	err_context;
};

union cs_res switch (cs_stat status) {
	case CS_OK:
		void;
	default:
		long	err_context;
};

typedef string volname<CS_MAXNAMLEN>;

struct cs_volargs {
	volname		vol;
};

struct cs_volokres {
	long	volume_id;
	/* verbatim copy from wfs.h */
	long	bytes_to_server;	/* # of bytes sent to servers */
	long	bytes_from_server;	/* # of bytes received from servers */
	long	bytes_to_client;	/* # of bytes sent to clients */
	long	bytes_from_client;	/* # of bytes received from clients */
	long	n_request_readdir;	/* # of readdir requests */
	long	n_request_lookup;	/* # of lookup requests */
	long	n_request_read;		/* # of read requests */
	long	n_request_readdir_miss;	/* # of readdir actually done */
	long	n_request_lookup_miss;	/* # of lookup actually done */
	long	n_request_read_miss;	/* # of read actually done */
};

union cs_volres switch (cs_stat status) {
	case CS_OK:
		cs_volokres volok;
	default:
		long	err_context;
};

struct cs_xferent {
	int	status;
	int	uid;		/* uid of originator */
	int	gid;		/* gid of originator */
	long	ip_address;	/* hostname of originator */
	long	bytes;		/* done so far */
	int	idle;		/* risky but no response */
	char	server<CS_MAXHOSTLEN>;	/* server name */
};

union cs_xferres switch (cs_stat status) {
	case CS_OK:
		cs_xferent xfer;
	default:
		long	err_context;
};

struct cs_udaargs {
	nfs_fh		file;
	string		attrname<CS_MAXNAMLEN>;
};

union cs_udares switch (cs_stat status) {
	case CS_OK:
		char	attr<CS_MAXDATA>;	/* attribute value */
	default:
		long	err_context;
};

struct cs_diropargs {
	nfspath		pname;
	filename	fname;
};

struct cs_diropokres {
	nfs_fh	file;
	fattr	attr;
};

union cs_diropres switch (cs_stat status) {
	case CS_OK:
		cs_diropokres diropok;
	default:
		long	err_context;
};

struct cs_readdirargs {
	nfs_fh		dir;
	long		cookie;
	unsigned	count;
};

struct cs_dirent {
	unsigned	fileid;
	filename	name;
	fattr		attr;
	cs_dirent	*nextentry;
};

struct cs_readdirokres {
	cs_dirent	*entries;
	bool		eof;
};

union cs_readdirres switch (cs_stat status) {
	case CS_OK:
		cs_readdirokres readdirok;
	default:
		long	err_context;
};

struct cs_readargs {
	nfs_fh		file;
	unsigned	offset;
	unsigned	count;
	unsigned	totalcount;
	bool		prefer_bulk;	/* prefer bulk data xfer */
};

struct cs_readokres {
	fattr	attr;
	opaque	data<CS_MAXDATA>;
};

struct cs_readbulkres {
	fattr	attr;
	long	sin_port;
};

union cs_readres switch (cs_stat status) {
	case CS_OK:
		cs_readokres readok;
	case CS_BULK:
		cs_readbulkres readbulk;
	default:
		long	err_context;
};

struct cs_geterrargs {
	long	err_context;
};

union cs_geterrres switch (cs_stat status) {
	case CS_OK:
		string err_string<CS_MAXNAMLEN>;
	default:
		void;
};

program CS_PROGRAM {
	version CS_VERSION {
		void
		CS_PROC_NULL(void) = 0;

		int
		CS_PROC_GETPORT(void) = 1;	/* obsolete */

		cs_fhres
		CS_PROC_REGISTER(void) = 2;

		cs_res
		CS_PROC_UNREGISTER(void) = 3;

		cs_res
		CS_PROC_MOUNT(cs_volargs) = 4;

		cs_res
		CS_PROC_UMOUNT(cs_volargs) = 5;

		cs_diropres
		CS_PROC_LOOKUP(cs_diropargs) = 6;

		cs_readdirres
		CS_PROC_READDIR(cs_readdirargs) = 7;

		cs_readres
		CS_PROC_READ(cs_readargs) = 8;

		cs_volres
		CS_PROC_GETVOL(cs_volargs) = 9;

		cs_xferres
		CS_PROC_GETXFER(nfs_fh) = 10;

		cs_udares
		CS_PROC_GETUDA(cs_udaargs) = 11;

		cs_geterrres
		CS_PROC_GETERR(cs_geterrargs) = 12;

	} = 1;
} = 300128;	/* XXX unlikely to be assigned */
