#!/usr/local/bin/perl
#
# Read $usg below.
# Author: Youki Kadobayashi <youki@wide.ad.jp>
# This file is part of WWFS.
#
require '/etc/csd.pl';

$usg = "makelist.pl: Make table of volumes from volume files.
Usage:	makelist [-i] [-o <output-file>] <volume-files...>
	-o	output to designated file (default: $WWFSDIR/vol/INDEX)
	-i	show volume id
";

require 'getopts.pl';
do Getopts('io:');
&checkup;

$: = " \n.-";

if ($#ARGV < $[) {
    die $usg;
}
if ($opt_o) {
    open(LIST, ">$opt_o")
	|| die "$opt_o: $!\n\n$usg";
} else {
    open(LIST, ">$WWFSDIR/vol/INDEX")
	|| die "$WWFSDIR/vol/INDEX: $!\n\n$usg";
}
foreach $file (@ARGV) {
    open(IN, "<$file");
    ($id, $volume, $server, $description) = ();	# clear
    $volume = &basename($file);
    $volume =~ s/.vol$//;
    while (<IN>) {
	if (/^volume-id:\s*(\S+)/i) {
	    $id = $1;
	};
	if (! $server && /^server:\s*(\S+)$/i) {
	    $server = $1;
	};
	if (! $server && /^ftp-server:\s*(\S+)$/i) {
	    $server = $1;
	};
	if (/^description:\s*(\S+.*)$/i) {
	    $description = $1;
	};
    }
    close(IN);
    if ($opt_i) {
	$volume = sprintf("%3d %s", $id, $volume);
    }
    write(LIST);
}
close(LIST);

format top =
volume             server                 description
------------------ ---------------------- -------------------------------------
.

format LIST =
^<<<<<<<<<<<<<<<<< ^<<<<<<<<<<<<<<<<<<<<< ^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$volume,           $server,               $description
^<<<<<<<<<<<<<<<<< ^<<<<<<<<<<<<<<<<<<<<< ^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$volume,           $server,               $description
.

sub checkup {
    if (!defined($WWFSDIR) || $WWFSDIR !~ m:^/:
	|| $WWFSDIR eq "" || $WWFSDIR eq "/") {
	die "/etc/csd.pl: \$WWFSDIR is not set correctly\n";
    }
}

sub basename {
    local ($pathname) = @_;
    local ($dir, $basename, $pos);

    $pos = rindex($pathname, "/");
    if ($pos < 0) {
        return $pathname;
    }
    ++$pos;
    ($dir, $basename) = unpack("a$pos a*", $pathname);
    return $basename;
}

# Local Variables:
# mode: perl
# End:
