#!/usr/local/bin/perl
#
# Read $usg below.
# Author: Youki Kadobayashi <youki@wide.ad.jp>
# This file is part of WWFS.
#
require '/etc/csd.pl';
$voldir = "$WWFSDIR/vol";

$usg = "map.pl: Check volume files and assign volume-id to them.
Usage:	map.pl <volume-files...>
	No option.
	$voldir -- for volumes files (*.vol).
";

if ($#ARGV < $[) {
    die $usg;
}

&checkup;

foreach $file (@ARGV) {
    if (&parse($file) == 0) {
	print "$file: OK\n";
    }
}
if (defined($seq)) {
    &writeseq;
}
exit 0;


# Check syntax of volume file.  Return value is
#	 0: if it's OK.
#	-1: if there's problem.
sub parse {
    local ($file) = @_;
    local ($id, @servers, @dirs);
    # parse the volume description file
    unless (open(FILE, "$file")) {
	print "$file: $!\n";
	return -1;
    }
    while (<FILE>) {
	/^volume-id:\s*(\d+)/i && do {
	    $id = $1;
	};
	/^ftp-server:\s*(\S+)$/i && do {
	    push(@servers, $1);
	};
	/^ftp-directory:\s*(\S+)$/i && do {
	    push(@dirs, $1);
	};
	/^server:\s*(\S+)$/i && do {
	    print "$file: warning: obsolete entry \"server:\".\n";
	    push(@servers, $1);
	};
	/^directory:\s*(\S+)$/i && do {
	    print "$file: warning: obsolete format \"directory:\".\n";
	    push(@dirs, $1);
	};
    }
    close(FILE);

    # check syntax
    if (!defined($id)) {
	if (!defined($seq)) {
	    &readseq;
	}
	if (open(FILE, ">> $file")) {
	    $id = $seq;
	    ++$seq;
	    print FILE "\nvolume-id: $id\n";
	    close(FILE);
	    print "$file: volume-id=$id was generated\n";
	} else {
	    print "$file: cannot append volume-id\n";  return -1;
	}
    }
    if (!defined(@servers)) {
	print "$file: no server\n";  return -1;
    }
    if (!defined(@dirs)) {
	print "$file: no directory\n";  return -1;
    }
    if ($#servers != $#dirs) {
	print "$file: server-directory mismatch\n";  return -1;
    }
    return 0;
}

sub checkup {
    if (!defined($WWFSDIR) || $WWFSDIR !~ m:^/:
	|| $WWFSDIR eq "" || $WWFSDIR eq "/") {
	die "/etc/csd.pl: \$WWFSDIR is not set correctly\n";
    }
    die "$voldir doesn't exist: install-server fail?\n" if (! -d $voldir);
}

sub readseq {
    if (open(SEQ, "< $voldir/seq")) {
	$seq = <SEQ>;
	chop $seq;
	if ($seq !~ /^\d+$/) {
	    print "fatal: seq file corrupt\n";
	    exit 1;
	}
	close(SEQ);
    } else {
	print "warning: cannot find seq file: starting volume-id from 1000\n";
	$seq = 1000;
    }
}

sub writeseq {
    if (open(SEQ, "> $voldir/seq")) {
	print SEQ "$seq\n";
	close(SEQ);
    }
}    

# Local Variables:
# mode: perl
# End:
