/* 
 * WorldWide File System
 * Copyright (c) 1992,1993 Youki Kadobayashi
 * Copyright (c) 1992,1993 Osaka University
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that the following conditions
 * are met:
 * 1. Both the copyright notice and this permission notice appear in
 *    all copies of the software, derivative works or modified versions,
 *    and any portions thereof, and that both notices appear in
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Osaka University
 *      and its contributors.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE DEVELOPER ``AS IS'' AND OSAKA
 * UNIVERSITY DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
 * WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Osaka University requests users of this software to return to
 *
 *  Youki Kadobayashi
 *  Department of Information and Computer Sciences
 *  Osaka University, Toyonaka 560, Osaka, Japan
 *
 * any improvements or extensions that they make and grant Osaka
 * University the rights to redistribute these changes.
 */
/* WWFS mount */
static char *AtFSid = "$Header: wwmount.c[109.0] Wed Nov 24 03:49:37 1993 youki-k@is.aist-nara.ac.jp saved $";

#include "am.h"		/* amd compatibility */

/* NFS headers and definitions */
#define NFS
#define NFSCLIENT
#ifdef NFS_3
typedef nfs_fh fhandle_t;
#endif /* NFS_3 */
#include <sys/mount.h>
#ifdef NFS_HDR
#include NFS_HDR
#endif /* NFS_HDR */

/* miscellaneous headers */
#include <sys/signal.h>
#include <sys/stat.h>
#include <string.h>			/* for strchr */
#include "cs_prot.h"

/* definition of behaviour */
#define	CSMNT_TIMEO	22
#define	CSMNT_RETRANS	5
#define	CSMNT_AC	8
#define	CSMNT_RETRY	10000
#define	BGRND	1
#define	ISBGRND	2
int opflags = 0;
#ifdef MOUNTED
char *mtab = MOUNTED;
#endif /* MOUNTED */
time_t clock_valid = 0;
int mypid;

void
usage(name)
char *name;
{
	fprintf(stderr, "Usage: %s server_host local_dir\n", name);
}

void
compute_mntent_options(mntp, fs_name, mp_name)
struct mntent *mntp;
char *fs_name, *mp_name;
{
	char *opts;
	opts = malloc(64);
	mntp->mnt_dir = mp_name;
	mntp->mnt_fsname = fs_name;
	mntp->mnt_type = MNTTYPE_WWFS;		/* or MTAB_TYPE_NFS; */
	sprintf(opts, "soft,timeo=%d,retrans=%d,acregmax=%d",
		CSMNT_TIMEO, CSMNT_RETRANS, CSMNT_AC);
	mntp->mnt_opts = opts;
	mntp->mnt_freq = 0;
	mntp->mnt_passno = 0;
}

void
compute_nfs_options(nfs_argsp)
struct nfs_args *nfs_argsp;
{
	nfs_argsp->flags = NFSMNT_SOFT | NFSMNT_HOSTNAME
		| NFSMNT_TIMEO | NFSMNT_RETRANS;
	nfs_argsp->timeo = CSMNT_TIMEO;
	nfs_argsp->retrans = CSMNT_RETRANS;
	nfs_argsp->rsize = nfs_argsp->wsize = NFS_MAXDATA;
#if defined(NFSMNT_NOAC) && !defined(sony_news)
	nfs_argsp->flags |= NFSMNT_NOAC;
#else /* NFSMNT_NOAC */
#if defined(NFSMNT_ACREGMIN) && defined(NFSMNT_ACREGMAX)
	nfs_argsp->flags |= NFSMNT_ACREGMIN | NFSMNT_ACREGMAX;
	nfs_argsp->acregmin = nfs_argsp->acregmax = CSMNT_AC;
	nfs_argsp->flags |= NFSMNT_ACDIRMIN | NFSMNT_ACDIRMAX;
	nfs_argsp->acdirmin = nfs_argsp->acdirmax = CSMNT_AC;
#endif
#endif /* NFSMNT_NOAC */
}

int
compute_mount_flags()
{
	int flags;
#ifdef NFS_4
	flags = M_NEWTYPE;
#else
	flags = 0;
#endif
	flags |= M_RDONLY | M_NOSUID;
	return flags;
}

char *
compute_fs_name(s)
char *s;
{
	char *fs_name;
	char *p;
	fs_name = malloc(strlen(s)+16);
	strcpy(fs_name, s);
#ifndef HOSTNAMESZ
#define SHORT_MOUNT_NAME
#endif /* HOSTNAMESZ */
#ifdef SHORT_MOUNT_NAME
	if (p = strchr(fs_name, '.'))
		*p = '\0';
#endif
#ifdef HOSTNAMESZ
	if (strlen(fs_name) >= HOSTNAMESZ)
		strcpy(fs_name + HOSTNAMESZ - 8, "..");
#endif
	if (strchr(fs_name, ':') == NULL) {
		strcat(fs_name, ":wwfs");
	}
	return fs_name;
}

/* code below is derived from NET2 mount.c */

int
compute_bindings(nfs_argsp, host, dir)
struct nfs_args *nfs_argsp;
char *host, *dir;
{
	int retry;
	struct hostent *hp;
	struct sockaddr_in *sinp = ALLOC(sockaddr_in);
	u_short nfs_port;
	cs_fhres *resp = ALLOC(cs_fhres);

	/* get 'sockaddr_in' of the file server */
	bzero(sinp, sizeof (*sinp));
	if ((hp = gethostbyname(host)) == NULL) {
		plog(XLOG_ERROR,
		     "mount: Can't get network address for host %s\n", host);
		return (-1);
	}
	bcopy(hp->h_addr, &(sinp->sin_addr), sizeof(sinp->sin_addr));

	/* try several times contacting server */
	retry = CSMNT_RETRY;
	resp->status = ETIMEDOUT;	/* Mark not yet successful */
	while (retry > 0) {
		nfs_port = contact_server(sinp, resp);
		if (nfs_port) {
			break;
		}
		if (--retry > 0) {
			if (opflags & BGRND) {
				opflags &= ~BGRND;
				if (fork())
					return (-1);
				else
					opflags |= ISBGRND;
			} 
			sleep(20);
		}
	}

	/* abort if error */
	if (resp->status) {
		if (opflags & ISBGRND)
			exit(1);
		plog(XLOG_ERROR, "Mount RPC error on %s", host);
		errno = resp->status;
		perror((char *)NULL);
		return (-1);
	}

	/* set nfs_argsp */
#if defined(__osf__) && defined(__alpha)
	sinp->sin_family = htons(AF_INET);	/* DEC OSF/1 bug? */
#else
	sinp->sin_family = AF_INET;
#endif
	sinp->sin_port = htons(nfs_port);
	NFS_SA_DREF(*nfs_argsp, sinp)
	NFS_FH_DREF(nfs_argsp->fh, (NFS_FH_TYPE) &resp->cs_fhres_u.file);
	nfs_argsp->hostname = host;

	return 0;
}

int
contact_server(sinp, resp)
struct sockaddr_in *sinp;
cs_fhres *resp;
{
	CLIENT *clntp;
	enum clnt_stat clnt_stat;
	struct timeval create_try, call_try;
	int so = RPC_ANYSOCK;
	int nfs_port;
	char fake_arg;

	sinp->sin_family = AF_INET;
	sinp->sin_port = htons(0);
	create_try.tv_sec = 10;
	create_try.tv_usec = 0;
	if ((clntp = clntudp_create(sinp, CS_PROGRAM, CS_VERSION,
				    create_try, &so)) == NULL) {
		if ((opflags & ISBGRND) == 0)
			clnt_pcreateerror("Failed to clntudp_create()");
		return 0;
	} else {
		clntp->cl_auth = authunix_create_default();
		/* try CS_PROC_REGISTER */
		call_try.tv_sec = 10;
		call_try.tv_usec = 0;
		clnt_stat = clnt_call(clntp, CS_PROC_REGISTER,
				      xdr_void, &fake_arg, /* in */
				      xdr_cs_fhres, resp, /* out */
				      call_try);
		/* ignore results for now */

		/* try CS_PROC_GETPORT */
		call_try.tv_sec = 10;
		call_try.tv_usec = 0;
		clnt_stat = clnt_call(clntp, CS_PROC_GETPORT,
				      xdr_void, &fake_arg, /* in */
				      xdr_int, &nfs_port, /* out */
				      call_try);
		if (clnt_stat != RPC_SUCCESS) {
			if ((opflags & ISBGRND) == 0)
				clnt_perror(clntp, "Failed to CS_PROC_REGISTER");
			return 0;
		} else {
			auth_destroy(clntp->cl_auth);
			clnt_destroy(clntp);
			return nfs_port;
		}
	}
}

int
mount_fs(mntp, flags, mnt_data, retry, type)
struct mntent *mntp;
int flags;
caddr_t mnt_data;	/* nfs_args actually */
int retry;
MTYPE_TYPE type;
{
	int error;
#ifdef MNTINFO_DEV
	struct stat stb;
#endif /* MNTINFO_DEV */

#ifdef FASCIST_DF_COMMAND
	/* Oh my SGI! */
	if (strcmp(mntp->mnt_type, MNTTYPE_WWFS) == 0) {
		mntp->mnt_type = FASCIST_DF_COMMAND;
	}
#endif /* FASCIST_DF_COMMAND */

	error = MOUNT_TRAP(type, mntp, flags, mnt_data);
	if (error < 0) {
		plog(XLOG_ERROR, "mount failed on %s: errno=%d", mntp->mnt_dir,
			errno);
		return (-1);
	}

#ifdef UPDATE_MTAB
#ifdef MNTINFO_DEV
	/*
	 * Add the extra dev= field to the mount table.
	 */
	if (lstat(mntp->mnt_dir, &stb) == 0) {
		char *zopts = (char *) malloc(strlen(mntp->mnt_opts) + 32);
		if (sizeof(stb.st_dev) == 2) {
			/* e.g. SunOS 4.1 */
			sprintf(zopts, "%s,%s=%s%04lx", mntp->mnt_opts,
				MNTINFO_DEV, MNTINFO_PREF,
				(u_long) stb.st_dev & 0xffff);
		} else {
			/* e.g. System Vr4 */
			sprintf(zopts, "%s,%s=%s%08lx", mntp->mnt_opts,
				MNTINFO_DEV, MNTINFO_PREF,
				(u_long) stb.st_dev);
		}
		mntp->mnt_opts = zopts;
	}
#endif /* MNTINFO_DEV */

#ifdef FIXUP_MNTENT
	/*
	 * Additional fields in struct mntent
	 * are fixed up here
	 */
	FIXUP_MNTENT(mntp);
#endif

	write_mntent(mntp);
#endif /* UPDATE_MTAB */

	return 0;
}

#ifdef MOUNT_HELPER_SOURCE
#include MOUNT_HELPER_SOURCE
#endif /* MOUNT_HELPER_SOURCE */

main(argc, argv)
int argc;
char **argv;
{
	struct nfs_args nfs_args;
	struct mntent mnt;
	int flags, ret;
	char *fs_name, *mp_name, *hostname;

	if (argc != 3) {
		usage(argv[0]);
		exit(1);
	}
	/* amd compat code */
	mypid = getpid();

	hostname = argv[1];
	mp_name = argv[2];
	fs_name = compute_fs_name(argv[1]);
	bzero(&nfs_args, sizeof (nfs_args));
	compute_nfs_options(&nfs_args);
	compute_mntent_options(&mnt, fs_name, mp_name);
	ret = compute_bindings(&nfs_args, hostname, mp_name);
	if (ret < 0) exit(1);

	flags = compute_mount_flags();
	ret = mount_fs(&mnt, flags, (caddr_t) &nfs_args, 1, MOUNT_TYPE_NFS);
	return ret;
}

